import numpy as np
from boss.utils.arrays import shape_consistent_XY


class UserFunc:
    """Wrapper class for BOSS user functions. """

    def __init__(self, func, dim):
        self.func = func
        self.dim = dim

    def eval(self, X_in, ygrad=False):
        X_in = np.atleast_2d(X_in)
        output = self.func(X_in)

        if isinstance(output, tuple):
            if not ygrad:
                X_out = output[0]
                Y = output[1]
            else:
                if len(output) > 2:
                    X_out = output[0]
                    Y = np.atleast_2d(output[1])
                    dY = np.atleast_2d(output[2])
                else:
                    X_out = X_in
                    Y = np.atleast_2d(output[0])
                    dY = np.atleast_2d(output[1])
                Y = np.concatenate((Y, dY), axis=1)
        else:
            Y = output
            X_out = X_in

        return shape_consistent_XY(X_out, Y, self.dim, ygrad=ygrad)

    def __call__(self, X_in):
        return self.eval(X_in)



import numpy as np
from scipy.spatial.distance import euclidean

import boss.io.ioutils as ioutils
import boss.io.parse as parse
from boss.bo.initmanager import InitManager
from boss.bo.kernel_factory import KernelFactory
from boss.bo.model import Model, GradientModel
from boss.bo.results import BOResults, minimize_model
from boss.bo.rstmanager import RstManager
from boss.io.main_output import MainOutput
from boss.settings import Settings
from boss.utils.arrays import shape_consistent_XY
from boss.utils.minimization import Minimization
from boss.utils.timer import Timer


class BOMain:
    """
    Class for handling Bayesian Optimization
    """

    def __init__(self, f, bounds, **keywords):
        keywords["bounds"] = bounds
        settings = Settings(keywords, f=f)
        self.settings = settings
        self.rst_manager = None
        self._setup()

    @classmethod
    def from_file(cls, ipfile, outfile=None, f=None, **new_keywords):
        """Initialize BOMain from a BOSS input or rst file.

        Parameters
        ----------
        ipfile : path_like
            The input file to initialize from, can be either
            a boss input or rst file.
        **new_keywords
            Any new BOSS keywords.
        """
        self = cls.__new__(cls)
        input_data = parse.parse_input_file(ipfile)
        rst_data = input_data.get("rst_data", np.array([]))
        keywords = input_data.get("keywords", {})
        keywords.update(new_keywords)
        self.settings = Settings(keywords, f=f)
        self.rst_manager = RstManager(self.settings, rst_data)
        cls._setup(self)
        return self

    @classmethod
    def from_settings(cls, settings, rst_data=None):
        """Construction from a Settings object."""
        self = cls.__new__(cls)
        self.settings = settings
        self.rst_manager = RstManager(self.settings, rst_data)
        self.restart = False
        cls._setup(self)
        return self

    def _setup(self):
        """Common setup for all factory methods."""
        settings = self.settings
        self.main_output = MainOutput(settings)
        if not self.rst_manager:
            self.rst_manager = RstManager(settings)

        self.init_manager = InitManager(
            settings["inittype"], settings["bounds"], settings["initpts"]
        )
        self.acqfn = settings.acqfn
        self.expfn = settings.expfn
        self.user_func = settings.f
        self.ygrad = settings['ygrad']

        self.kernel = KernelFactory.construct_kernel(settings)
        self.model = None
        Minimization.set_parallel(self.settings["parallel_optims"])

        self.results = BOResults(settings=settings)
        self.results.add_defaults()
        self.itr_curr = 0

    def init_model(self, X, Y, params=None):
        """Initializes the GP model."""
        X, Y = shape_consistent_XY(
            X, Y, self.settings.dim, nan_pad=False, ygrad=self.ygrad
        )
        model = Model if not self.ygrad else GradientModel
        self.model = model(
            X,
            Y,
            self.kernel,
            self.settings["noise"],
            self.settings["ynorm"],
        )
        # Look for optimized model params from:
        # func arguments -> restart manager -> optimize model
        if params is None:
            num_params = self.model.get_unfixed_params().size
            params = self.rst_manager.get_theta(X.shape[0] + 1, num_params)
        if not np.any(params == None):
            self.model.set_unfixed_params(params)
        else:
            if self.settings["initupdate"]:
                self.model.optimize(self.settings["updaterestarts"])
        self.acqfn.model = self.model
        self.expfn.model = self.model

    def init_run(self, X_init, Y_init=None):
        """Evalutes initial points and calls model initialization.

        This method assumes that, at least, initial X-values have been determined,
        i.e. passed by the user or retrieved from file or an InitManager
        (see resolve_initpts).

        Parameters
        ----------
        X_init: np.ndarray
            Initial X-values, must be completely specified,
            i.e., cannot be None nor have nan-elements.

        Y_init: Optional[np.ndarray]
            Initial Y-values, can be omitted entirely or partially, in which
            case the user function will be evaluated at the missing X-locations.

        Returns
        -------
        X_next: np.ndarray
            The very first acquisition that will be used in the BO-run.
        """
        self.main_output.new_file()
        self.rst_manager.new_file()
        self.results.clear()

        # Ensure X,Y initial values are 2D arrays of the same size.
        dim = self.settings.dim
        X_init, Y_init = shape_consistent_XY(
            X_init, Y_init, dim, nan_pad=True, ygrad=self.ygrad
        )
        self.settings["initpts"] = X_init.shape[0]

        X = np.empty((0, dim), float)
        Y = np.empty((0, 1 + self.ygrad * self.settings.dim), float)
        for i in range(self.settings["initpts"]):
            with self.main_output.summarize_results(self.results):

                # Evaluate the userfn if initial y-data is not available
                if np.isnan(Y_init[i]).all():
                    XY_out = self._eval_user_func(X_init[i, :])
                    X = np.vstack((X, XY_out[0]))
                    Y = np.vstack((Y, XY_out[1]))
                else:
                    X = np.vstack((X, X_init[i, :]))
                    Y = np.vstack((Y, Y_init[i, :]))
                    self.rst_manager.new_data(X_init[i, :], Y_init[i, :])

                # For the first n-1 initpts the only results are (x,y)-data.
                if i < self.settings["initpts"] - 1:
                    self.results.update({"X": X, "Y": Y})
                # For the final initpt: initialize model and calc. all results.
                else:
                    self.init_model(X, Y)

                    # Add model parameters to rst-file.
                    self.rst_manager.new_model_params(self.model.get_unfixed_params())

                    # Get the next acquisition and update all results.
                    X_next = self.acquire()
                    self._update_results(X_next)

        self.results.set_num_init_batches(self.settings["initpts"])
        return X_next

    def get_initpts(self):
        """If initial data is not provided, get it from the rst/init manager."""
        if self.rst_manager.data.shape[0] > 0:
            X_init = self.rst_manager.X
            Y_init = self.rst_manager.Y
        else:
            X_init = self.init_manager.get_all()
            Y_init = np.empty(
                (len(X_init), 1 + self.ygrad * self.settings.dim)) * np.nan
        return X_init, Y_init

    def run(self, X_init=None, Y_init=None, iterpts=None):
        """
        The Bayesian optimization main loop. Evaluates first the initialization
        points, then creates a GP model and uses it and an acquisition function
        to locate the next points where to evaluate. Stops when a pre-specified
        number of initialization points and BO points have been acquired or a
        convergence criterion is met.

        Returns
        -------
        BOResults
            An object that provides convenient access to the most
            important results from the optimization.
        """
        if iterpts:
            self.settings["iterpts"] = iterpts
        # Initialize model & files if we're starting fresh or restarting from file.
        if self.model is None:
            # Handle initial points
            if X_init is not None:
                X_init, Y_init = shape_consistent_XY(
                    X_init,
                    Y_init,
                    self.settings.dim,
                    nan_pad=True,
                    ygrad=self.ygrad,
                )
            else:
                X_init, Y_init = self.get_initpts()

            X_next = self.init_run(X_init, Y_init)
        # The model already exists and we just need the next acquisition.
        else:
            X_next = self.acquire()

        # BO main loop
        for _ in range(self.itr_curr, self.settings["iterpts"] + 1):
            with self.main_output.summarize_results(self.results):

                # 1. User func evaluation
                X_new, Y_new = self._eval_user_func(X_next)

                # 2. Model update: refit model to new data & optimize params.
                self._update_model(X_new, Y_new)

                # 3. Get a new acquisition for the next iteration.
                X_next = self.acquire()

                # 4. Update results and write a summary to the outfile.
                self._update_results(X_next)

            # 5. Convergence check
            if self.has_converged():
                self.main_output.convergence_stop()
                break

        return self.results

    def _eval_user_func(self, X):
        """Evalutes the userfn and writes data to the rstfile."""
        local_timer = Timer()

        # Evaluate the user function
        with local_timer.time():
            X_out, Y = self.user_func.eval(X, self.ygrad)

        # Write info to file
        for x, y in zip(X_out, Y):
            self.rst_manager.new_data(x, y)
            self.main_output.progress_msg(
                "Evaluating objective function at x ="
                + ioutils.oneDarray_line(x, len(x), float),
            )
        self.main_output.progress_msg(
            "Objective function evaluated," + f" time [s] {local_timer.lap_time}",
        )
        return X_out, Y

    def _update_model(self, X_next, Y_next):
        """Optimizes and refits model with new data."""

        # Store new data
        self.model.add_data(X_next, Y_next)

        # Optimize model if needed.
        updatefreq = self.settings["updatefreq"]
        updaterestarts = self.settings["updaterestarts"]
        updateoffset = self.settings["updateoffset"]

        itr1 = self.itr_curr - 1  # start optimizing on iteration 1
        should_optimize = (
            updatefreq > 0 and itr1 >= updateoffset and (itr1 % updatefreq == 0)
        )
        if should_optimize:
            self.model.optimize(updaterestarts)

        # Add model parameters to rst-file.
        self.rst_manager.new_model_params(self.model.get_unfixed_params())

    def _update_results(self, X_next):
        """Updates BOResults with acqs, model and min info."""
        minfreq = self.settings["minfreq"]
        self.results.update(
            {
                "X": self.model.X,
                "Y": self.model.Y,
                "X_next": X_next,
                "model_params": self.model.get_unfixed_params(),
            }
        )
        # Add global min. info
        should_optimize = self.itr_curr == self.settings["iterpts"] or (
            minfreq > 0 and (self.itr_curr % minfreq == 0)
        )

        if should_optimize:
            x_glmin, mu_glmin, nu_glmin = minimize_model(
                self.model,
                self.settings["bounds"],
                self.settings["optimtype"],
                self.settings["kernel"],
                self.settings["min_dist_acqs"],
                accuracy=self.settings["minzacc"],
            )
            self.results.update(
                {
                    "x_glmin": x_glmin,
                    "mu_glmin": mu_glmin,
                    "nu_glmin": nu_glmin,
                }
            )
        self.itr_curr += 1

    def has_converged(self):
        """
        Checks whether dxhat has been within tolerance for long enough
        TODO: should use dxmuhat instead?
        """
        glmin_tol = self.settings["glmin_tol"]
        if glmin_tol is None:
            return False

        conv_tol, conv_itrs = glmin_tol
        minfreq = self.settings["minfreq"]
        num_lookback = max(minfreq, conv_itrs) // minfreq
        x_glmin = self.results["x_glmin"]
        if len(x_glmin.data) < num_lookback:
            return False

        for i in range(1, num_lookback + 1):
            dx = euclidean(x_glmin.value(-i), x_glmin.value(-i - 1))
            if dx > conv_tol:
                return False

        return True

    def acquire(self):
        """
        Get a new point to evaluate by either reading it from the rst-file or,
        in case it doesn't contain the next point to evaluate, by obtaining
        a new initial point (when run is in initialization stage) or
        minimizing the acquisition function (when the run is in BO stage).
        """
        minimizer_params = {
            "bounds": self.settings["bounds"],
            "optimtype": self.settings["optimtype"],
        }
        x_next = self.acqfn.minimize(**minimizer_params)
        if self._loc_overconfident(x_next):
            x_next = self.expfn.minimize(**minimizer_params)
            return x_next
        return x_next

    def _loc_overconfident(self, x_next):
        """
        Checks is model variance is lower than tolerance at suggested X_next.
        """
        acqtol = self.settings["acqtol"] if self.acqfn is not self.expfn else None
        if acqtol is None:
            return False
        else:
            var_next = self.model.predict(x_next)[1]
            if var_next >= acqtol**2:
                return False
            else:
                output_msg = (
                    "Acquisition location " + "too confident, doing pure exploration"
                )
                self.main_output.progress_msg(output_msg)
                return True



import os
import shutil
import warnings

import numpy as np

import boss.io.dump as dump
import boss.io.ioutils as ioutils
import boss.pp.plot as plot
from boss.bo.results import BOResults
from boss.io.main_output import MainOutput
from boss.utils.minimization import Minimization

warnings.filterwarnings("ignore")  # ignore warnings


class PPMain:
    """Performs the automated post-processing of a BOSS run."""

    def __init__(
        self,
        bo_results,
        pp_iters=None,
        pp_acq_funcs=None,
        pp_models=None,
        pp_model_slice=None,
        pp_var_defaults=None,
        pp_truef_at_glmins=None,
        pp_truef_npts=None,
        pp_local_minima=None,
    ):
        pp_settings = {
            "pp_iters": pp_iters,
            "pp_acq_funcs": pp_acq_funcs,
            "pp_models": pp_models,
            "pp_model_slice": pp_model_slice,
            "pp_var_defaults": pp_var_defaults,
            "pp_truef_at_glmins": pp_truef_at_glmins,
            "pp_truef_npts": pp_truef_npts,
            "pp_local_minima": pp_local_minima,
        }
        self.results = bo_results
        self.settings = bo_results.settings
        self.settings.update({k: v for k, v in pp_settings.items() if v is not None})
        self.settings.correct()

        self.rstfile = self.settings["rstfile"]
        self.outfile = self.settings["outfile"]
        self.main_output = None
        self._setup()

    @classmethod
    def from_file(cls, rstfile, outfile, main_output=None):
        self = cls.__new__(cls)
        self.results = BOResults.from_file(rstfile, outfile)
        self.settings = self.results.settings
        self.rstfile = rstfile
        self.outfile = outfile
        self.main_output = main_output
        cls._setup(self)
        return self

    def _setup(self):
        """Common setup for all factory methods."""

        self.batch_tracker = self.results.batch_tracker

        if self.settings["pp_iters"] is None:
            self.pp_iters = np.arange(0, self.settings["iterpts"] + 1)
        else:
            self.pp_iters = self.settings['pp_iters']

        # User-specified model slices are interpreted as being indexed
        # from 1, so we convert to 0-based indexing here.
        self.settings["pp_model_slice"][:2] -= 1

        if not getattr(self, "main_output", None):
            self.main_output = MainOutput(self.settings)

        # Store paths to all PP-related files.
        self.files = {
            "acqs": "postprocessing/acquisitions.dat",
            "min_preds": "postprocessing/minimum_predictions.dat",
            "conv_meas": "postprocessing/convergence_measures.dat",
            "hypers": "postprocessing/hyperparameters.dat",
            "truef_glmin": "postprocessing/true_f_at_x_glmin.dat",
        }

        # Define a dict to hold all data structures used to generate the PP output.
        # The fundamental items are acqs, mod_par and min_preds: these must be extracted
        # from a BOResults object or read from file and can the be used to generate
        # xnexts, est_yranges and conv_meas.
        self.data = {
            "acqs": None,
            "mod_par": None,
            "min_preds": None,
            "xnexts": None,
            "est_yranges": None,
            "conv_meas": None,
        }

    def create_dirs(self):
        # Create required directories.
        if os.path.isdir("postprocessing"):
            print("warning: overwriting directory 'postprocessing'")

        shutil.rmtree("postprocessing", ignore_errors=True)
        os.makedirs("postprocessing", exist_ok=True)

        if self.settings["pp_models"]:
            os.makedirs("postprocessing/data_models", exist_ok=True)
            os.makedirs("postprocessing/graphs_models", exist_ok=True)
        if self.settings["pp_acq_funcs"]:
            os.makedirs("postprocessing/data_acqfns", exist_ok=True)
            os.makedirs("postprocessing/graphs_acqfns", exist_ok=True)
        if self.settings["pp_local_minima"] is not None:
            os.makedirs("postprocessing/data_local_minima", exist_ok=True)

    @property
    def slc_dim(self):
        sts = self.settings
        _slc_dim = 1 if sts["pp_model_slice"][0] == sts["pp_model_slice"][1] else 2
        return _slc_dim

    def run(self):
        self.load_data()
        self.dump_data()
        self.plot()

    def load_data(self):
        sts = self.settings
        res = self.results

        # calculate missing minimum info, if any
        # -> dont have to care about minfreq
        res.calc_missing_minima()

        # Get basic output data (acqs, mod_bar, min_preds) from BOResults
        iter_labels = res.batch_tracker.iteration_labels
        acqs = np.c_[iter_labels, res["X"], res["Y"]]

        ipars = np.fromiter(res["model_params"].keys(), dtype=int)
        mod_par = np.c_[ipars, res["model_params"]]

        imins = np.fromiter(res["mu_glmin"].keys(), dtype=int)
        min_preds = np.c_[
            imins,
            res["x_glmin"].to_array(),
            res["mu_glmin"].to_array(),
            res["nu_glmin"].to_array(),
        ]
        # Generate the derived output data: xnexts, conv_meas and est_yranges.
        xnexts = np.c_[iter_labels[:-1], acqs[1:, 1:sts.dim + 1]]

        est_yranges = np.c_[
            iter_labels,
            np.maximum.accumulate(acqs[:, -1]) - np.minimum.accumulate(acqs[:, -1]),
        ]

        conv_meas = None
        imins1 = min_preds[1:, 0].astype(int)
        yranges_at_imins = np.diff(np.array([res.get_est_yrange(i) for i in imins1]))
        conv_meas = np.c_[
            imins1,
            np.linalg.norm(min_preds[1:, 1:-2] - min_preds[:-1, 1:-2], axis=1),
            # np.abs(np.diff(min_preds[:, -2])) / est_yranges[initpts:, -1]
            np.abs(np.diff(min_preds[:, -2])) / yranges_at_imins,
        ]

        self.data.update(
            {
                "conv_meas": conv_meas,
                "est_yranges": est_yranges,
                "xnexts": xnexts,
                "acqs": acqs,
                "mod_par": mod_par,
                "min_preds": min_preds,
            }
        )

    def dump_data(self):
        """Dump standard data."""
        self.create_dirs()
        acqs = self.data["acqs"]
        conv_meas = self.data["conv_meas"]

        # Initialize dump files
        # 1. acquisitions
        ioutils.overwrite(
            self.files["acqs"],
            "# Data acquisitions " + "by iteration (iter npts x y)\n",
        )

        # 2. Minimum predictions
        ioutils.overwrite(
            self.files["min_preds"],
            "# Global minimum predictions by iteration"
            + " (iter npts x_glmin mu_glmin nu_glmin)\n",
        )

        # 3. Convergence measures
        if conv_meas is not None:
            ioutils.overwrite(
                self.files["conv_meas"],
                "# Convergence measures by iteration"
                + " (iter npts dx_glmin abs(dmu_glmin)/yrange)\n",
            )

        # 4. Hyperparameters
        ioutils.overwrite(
            self.files["hypers"],
            "# Model hyperparameter values by iteration" + " (iter npts parameters)\n",
        )
        # 5. True function values at xhats
        if self.settings["pp_truef_at_glmins"]:
            ioutils.overwrite(
                self.files["truef_glmin"],
                "# True function value at x_glmin locations"
                + " by iteration (iter npts f(x_glmin) "
                + "f(x_glmin)-mu_glmin)\n",
            )

        # dump acqs separately
        for i, acq in enumerate(acqs):
            it = acq[0]
            npts = i + 1
            ioutils.append_write(
                self.files["acqs"],
                ioutils.data_line([it, npts], acq[1:], fstr="%18.10E"),
            )

        # --- dump per-iteration quantities
        for it in range(self.batch_tracker.num_iters):
            npts = self.batch_tracker.ensemble_sizes[it]

            # hyperparameters
            par = self.results.select("model_params", it)
            if par is not None:
                ioutils.append_write(
                    self.files["hypers"],
                    ioutils.data_line([it, npts], par, fstr="%18.10E"),
                )

            # global minimum predictions
            mu_glmin = self.results.select("mu_glmin", it)
            x_glmin = self.results.select("x_glmin", it)
            if ~np.isnan(mu_glmin) and x_glmin is not None:
                nu_glmin = self.results.select("nu_glmin", it)
                glmin = np.append(x_glmin, [mu_glmin, nu_glmin])
                ioutils.append_write(
                    self.files["min_preds"],
                    ioutils.data_line([it, npts], glmin, fstr="%18.10E"),
                )

            # convergence measures
            if it in conv_meas[:, 0].astype(int):
                ind = list(conv_meas[:, 0].astype(int)).index(it)
                ioutils.append_write(
                    self.files["conv_meas"],
                    ioutils.data_line(
                        [
                            it,
                            npts,
                            conv_meas[ind, -2],
                            conv_meas[ind, -1],
                        ],
                        fstr="%18.10E",
                    ),
                )

            # true function at glmins
            if self.settings["pp_truef_at_glmins"]:
                self.main_output.progress_msg("Evaluating true function at x_glmin")
                X_glmin, f_glmin = self.settings.f.eval(np.atleast_2d(x_glmin))
                f_glmin = f_glmin[0].item()
                # ind = np.where(np.isclose(X_glmin, x_glmin).all(axis=1))[0].item()
                # f_glmin = f_glmin[ind].item()
                
                ioutils.append_write(
                    self.files["truef_glmin"],
                    ioutils.data_line(
                        [it, npts, f_glmin, f_glmin - mu_glmin], fstr="%18.10E"
                    ),
                )

    def plot(self, target="all"):
        if target in ["hyperparameters", "all"]:
            self._plot_hypers()

        if target in ["acquisitions", "all"]:
            self._plot_acqs()
        if target in ["convergence", "all"]:
            self._plot_conv()
        if target in ["model", "all"]:
            self._plot_model()
        if target in ["truef", "all"]:
            self._plot_truef()

    def _plot_hypers(self):
        # plot standard data dumps
        hypers = np.atleast_2d(np.loadtxt(self.files["hypers"], skiprows=1))
        if len(hypers) > 1:
            plot.plot_hyperparameters(
                self.settings, "postprocessing/hyperparameters.png", hypers
            )

    def _plot_acqs(self):
        minp = np.atleast_2d(np.loadtxt(self.files["min_preds"], skiprows=1))

        if len(minp[0]) == 0:
            raise ValueError("No minimal points found in out-file.")
        acqs_data = np.atleast_2d(np.loadtxt(self.files["acqs"], skiprows=1))

        if len(acqs_data) > 1:
            plot.plot_data_acquisitions(
                self.settings,
                "postprocessing/acquisition" + "_locations.png",
                acqs_data,
                minp,
            )

    def _plot_conv(self):
        minp = np.atleast_2d(np.loadtxt(self.files["min_preds"], skiprows=1))

        if len(minp[0]) == 0:
            raise ValueError("No minimal points found in out-file.")

        conv_meas = np.atleast_2d(np.loadtxt(self.files["conv_meas"], skiprows=1))
        if len(conv_meas) > 1:
            plot.plot_conv_measures(
                self.settings, "postprocessing/convergence_measures.png", conv_meas
            )

        if self.settings["pp_truef_at_glmins"]:
            truef_hats = np.atleast_2d(
                np.loadtxt(self.files["truef_glmin"], skiprows=1)
            )
            if len(truef_hats) > 1:
                plot.plot_truef_hat(
                    self.settings,
                    "postprocessing/true_function" + "_at_x_glmin.png",
                    truef_hats,
                )

    def _plot_model(self):
        sts = self.settings
        slc_dim = self.slc_dim

        # recreate snapshots of model in a loop for all pp_iters
        curr_xhat = None

        for it in self.pp_iters:

            npts = self.batch_tracker.ensemble_sizes[it]

            self.main_output.progress_msg("Post-processing iteration %i" % (it))
            model = self.results.reconstruct_model(it)
            assert npts == model.X.shape[0], "Model recreate fail!"

            # find current xhat
            curr_xhat = self.results.select("x_glmin", it)

            # Local minima
            if sts["pp_local_minima"] is not None:
                self.main_output.progress_msg("Finding model local minima")
                mins = Minimization.minimize(
                    model.predict_mean_grad,
                    sts["bounds"],
                    sts["kernel"],
                    np.hstack([model.X, model.Y]),
                    sts["min_dist_acqs"],
                    accuracy=sts["pp_local_minima"],
                    args=(),
                    lowest_min_only=False,
                )

                mins = sorted(mins, key=lambda x: (x[1]))
                minima_data = [[]] * (sts.dim + 2)
                for m in mins:
                    p = []
                    for i in range(sts.dim):
                        p.append(m[0][i])
                    mu, var = model.predict(m[0])
                    nu = np.sqrt(var)
                    p.append(mu)
                    p.append(nu)
                    minima_data = np.insert(minima_data, len(minima_data[0]), p, axis=1)

                titleLine = "# Local minima (x mu nu) - model data ensemble size %i" % (
                    npts
                )
                ioutils.write_cols(
                    "postprocessing/data_local_minima/"
                    "it%.4i_npts%.4i.dat" % (it, npts),
                    minima_data,
                    titleLine=titleLine,
                )
            # Model (cross-sections)
            if sts["pp_models"]:
                dump.dump_model(
                    sts,
                    "postprocessing/data_models/it%.4i_npts%.4i.dat" % (it, npts),
                    model,
                    model.get_all_params(),
                    curr_xhat,
                )
                mdata = np.loadtxt(
                    "postprocessing/data_models/" + "it%.4i_npts%.4i.dat" % (it, npts),
                    skiprows=2,
                )

                macqs = np.hstack(
                    (
                        self.results.select("X", np.arange(0, it + 1)),
                        self.results.select("Y", np.arange(0, it + 1)),
                    )
                )

                xnext = self.results.get_next_acq(it)

                if sts["pp_local_minima"] is not None and slc_dim == sts.dim:
                    minima = np.atleast_2d(np.array(minima_data)).T
                else:
                    minima = None

                plot.plot_model(
                    sts,
                    "postprocessing/graphs_models/it%.4i_npts%.4i" ".png" % (it, npts),
                    mdata,
                    curr_xhat,
                    macqs,
                    xnext,
                    minima,
                )

            # acquisition function (cross-sections)
            if sts["pp_acq_funcs"]:
                if xnext is not None:
                    if len(xnext.shape) == 2:
                        defs = xnext[0, :]
                    else:
                        defs = xnext
                else:
                    defs = curr_xhat
                acqfn = self.results.reconstruct_acq_func(it)
                dump.dump_acqfn(
                    sts,
                    "postprocessing/data_acqfns/it%.4i_npts%.4i.dat" % (it, npts),
                    acqfn,
                    defs,
                )
                acqfn_data = np.loadtxt(
                    "postprocessing/data_acqfns/" + "it%.4i_npts%.4i.dat" % (it, npts),
                    skiprows=1,
                )

                plot.plot_acq_func(
                    sts,
                    "postprocessing/graphs_acqfns/" "it%.4i_npts%.4i.png" % (it, npts),
                    acqfn_data,
                    macqs,
                    curr_xhat,
                    xnext,
                )

    def _plot_truef(self):
        """Dump and plot true function (cross-section)."""

        sts = self.settings
        min_preds = self.data["min_preds"]
        acqs = self.data["acqs"]
        xnexts = self.data["xnexts"]
        slc_dim = self.slc_dim
        res = self.results

        if sts["pp_truef_npts"] is None:
            return

        self.main_output.progress_msg("Dumping and plotting true function")

        curr_xhat = min_preds[-1, 1 : sts.dim + 1]

        dump.dump_truef(sts, "postprocessing/true_func.dat", curr_xhat)
        truef_data = np.loadtxt("postprocessing" + "/true_func.dat", skiprows=1)
        plot.plot_truef(sts, "postprocessing/true_func.png", truef_data)
        ind = np.where(min_preds[:, 1 : sts.dim + 1] == curr_xhat)[0][0]
        truef_slc_xhat_npts = int(min_preds[ind, 0])

        # replot 1D models with truef if it is now available
        if not (
            sts["pp_model_slice"][0] == sts["pp_model_slice"][1] and sts["pp_models"]
        ):
            print(
                "Replotting of 1D models with the true function requires pp_models and pp_model_slice[0]==pp_model_slice[1]"
            )
            return

        self.main_output.progress_msg("Replotting 1D models with true" + " function")
        for mdat_file in os.listdir("postprocessing/data_models"):

            # find it and and npts from naming it%.4i_npts%.4i.dat
            negit = True if mdat_file[2] == "-" else False
            it = int(mdat_file[2:6]) if not negit else int(mdat_file[2:7])
            npts = int(mdat_file[-8:-4])
            mdata = np.loadtxt(
                "postprocessing/data_models/" + "it%.4i_npts%.4i.dat" % (it, npts),
                skiprows=2,
            )
            xhat = res.select("x_glmin", it)

            macqs = np.hstack(
                (
                    self.results.select("X", np.arange(0, it + 1)),
                    self.results.select("Y", np.arange(0, it + 1)),
                )
            )
            # if slc_dim < sts.dim and slc_dim == 1:
            #     macqs = None
            xnext = self.results.get_next_acq(it)

            if sts["pp_local_minima"] is not None and slc_dim == sts.dim:
                minima = np.loadtxt(
                    "postprocessing/data_local_"
                    + "minima/it%.4i_npts%.4i.dat" % (it, npts),
                    skiprows=1,
                )
                minima = np.atleast_2d(minima)
            else:
                minima = None

            if npts != truef_slc_xhat_npts and slc_dim < sts.dim:
                truef_d = None
            else:
                truef_d = truef_data

            plot.plot_model(
                sts,
                "postprocessing/graphs_models/" + "it%.4i_npts%.4i.png" % (it, npts),
                mdata,
                xhat,
                macqs,
                xnext,
                minima,
                truef_d,
            )



from typing import *
import functools

import numpy as np

from boss.utils.minimization import Minimization
from boss.bo.kernel_factory import KernelFactory
from boss.bo.model import Model
from boss.bo.acq.explore import Explore
from boss.settings import Settings
from boss.utils.sparselist import SparseList
from boss.utils.sparselist import wrap_index, wrap_index_array
from boss.bo.acq.base import BaseAcquisition
import boss.io.parse as parse
import copy


def minimize_model(
    model,
    bounds,
    optimtype='score',
    kernel=None,
    min_dist_acqs=None,
    accuracy=0.3,
) -> Tuple[np.ndarray, float, float]:
    """Minimizes a surrogate model to provide a global minimum prediciton.

    Parameters
    ----------
    model : Model
        A give BOSS model.

    Returns
    -------
    Tuple[np.ndarray, float, float]:
        The global minimum prediction as a tuple 
        (min. value, min. location, min. variance).
    """
    if optimtype == 'score':
            gmin = Minimization.minimize_using_score(
                model.predict_mean_grad, bounds
            )
    else:
        gmin = Minimization.minimize(
            func=model.predict_mean_grad,
            bounds=bounds,
            kerntype=kernel,
            acqs=np.hstack([model.X, model.Y]),
            min_dist_acqs=min_dist_acqs,
            accuracy=accuracy,
        )

    x_glmin = np.array(gmin[0])
    mu_glmin, nu_glmin = model.predict(np.atleast_2d(x_glmin))
    return x_glmin, mu_glmin[0, 0], nu_glmin[0, 0]


class BatchTracker:
    """Tracks and handles the indices of all data batches.

    This object essentially stores the index ranges of all batches, i.e.,
    the indices of the first and last points for each batch in X. These
    indices can then be used to easily extract selected batches of X, Y etc.
    """
    def __init__(self, num_init_batches):
        """Initializes the batch tracker object.

        The the ranges of each batch are stored in the ind_lims variable
        such that X[ind_lims[i-1]:ind_lims[i]] yields the batch corresponding
        to iteration i. This indexing is then offset by the number of initial 
        batches, i.e. the number of batches that comprise the 0th iteration.

        Parameters
        ----------
        num_init_batches : int
            The number of batches that make up the initial data points (e.g., Sobol).

        """
        self._ind_lims = [0]
        self.num_init_batches = num_init_batches

    def update(self, num_data):
        """Updates the absolute index limits for new batches.
        
        The upper batch limit is given by the total number of 
        data points in X, i.e. this method is typically called 
        with num_data = len(X).
        """
        self._ind_lims.append(num_data)

    def get_batch_indices(self, itr):
        """Determines the batch indices for a given sequence of iterations.

        Batch indices I for a batch B are defined s.t. X[I, :] yields all
        x-data points in batch B.

        Parameters
        ----------
        itr : int
            The iteration for which we would like to obtain batch indices.

        Returns
        -------
        slice
            Batch indices corresponding to itr.
        """
        if itr == 0:
            inds = slice(0, self._ind_lims[self.num_init_batches])
        else:
            start = self._ind_lims[self.num_init_batches + itr - 1]
            stop = self._ind_lims[self.num_init_batches + itr]
            inds = slice(start, stop)
        return inds

    @property
    def num_iters(self):
        """Returns the number of iterations.

        Note that this number is equal to iterpts + 1 since we count
        the initial data as iteration 0.
        """
        return len(self._ind_lims) - self.num_init_batches

    def get_multibatch_indices(self, iters):
        """Determines the batch indices for a given sequence of iterations.

        Extends get_batch_indices to handle multiple iterations.
        
        Parameters
        ----------
        iters : Sequence
            A sequence of iteration numbers for which we would
            like to obtain the coresponding batch indices.

        Returns
        -------
        slice | np.ndarray
            Batch indices corresponding to iters.
        """
        # Single iteration
        if np.ndim(iters) == 0:
            inds = self.get_batch_indices(iters)
        # Increasing sequence of iterations without gaps
        elif np.all(np.diff(iters) == 1):
            start = self.get_batch_indices(np.min(iters)).start
            stop = self.get_batch_indices(np.max(iters)).stop
            inds = slice(start, stop)
        else: 
            inds = []
            for itr in iters:
                slc = self.get_batch_indices(itr) 
                inds.append(np.arange(slc.start, slc.stop))
            inds = np.hstack(inds)
        return inds

    @property
    def iteration_labels(self):
        """Returns the batch number of each data point.  """
        n = self.num_init_batches 
        reps = np.diff(self._ind_lims)
        reps = np.insert(reps[n:], 0, np.sum(reps[:n])) 
        labels = np.repeat(np.arange(self.num_iters), reps)
        return labels

    @property
    def batch_sizes(self):
        """Returns the batch size for each iteration. """
        n = self.num_init_batches 
        size_init = np.sum(np.diff(self._ind_lims[:n + 1]))
        sizes = np.insert(np.diff(self._ind_lims[n:]), 0, size_init) 
        return sizes

    @property
    def ensemble_sizes(self):
        """Returns the total size of the dataset for each iteration. """
        return np.cumsum(self.batch_sizes)


class BOResults:
    """Stores and handles results from the BO.

    This object essentially stores and updates raw result arrays/lists
    in the data dict, then provides easy access to them through the
    select method.

    The following terminology is used in this class:

    - An extendable result is appended to throughout the BO, .e.g, the collection
    of predicted global minima. These results are managed and updated directly by
    the Results object. Conversely, a non-extendable result, such as the
    acquisitions X, is owned and updated by another object
    (the model class in case of X). For non-extendable results the BOResults
    object just keeps an up-to-date reference to the object.

    - A sparse result is a result that is not necessarily recorded every iteration,
    e.g., the collection of global minima (which is both extendable and sparse).
    In their raw form, these results are stored using the SparseList class.
    """
    def __init__(self, settings: Settings) -> None:
        self.data = {}
        self.extendable_names = []
        self.settings = settings
        self.batch_tracker = BatchTracker(num_init_batches=1)

    def add_defaults(self):
        """Adds predefined data structures for the most common results. """
        for name in ["X", "Y", "X_next"]:
            self.add_new(name, extendable=False)
        for name in ["mu_glmin", "nu_glmin"]:
            self.add_new(name, extendable=True, sparse=True, default=np.nan)
        for name in ['x_glmin', 'model_params']:
            self.add_new(name, extendable=True, sparse=True, default=None)

    def __getitem__(self, name: str) -> Any:
        return self.data[name]

    def __setitem__(self, name: str, val) -> None:
        self.data[name] = val

    def get(self, name: int, default: Optional[Any] = None) -> Any:
        return self.data.get(name, default)

    def clear(self):
        """Clears all the data from the results.

        The name persist and are reverted to their initial state.
        Also attaches a new batch tracker.
        """
        self.batch_tracker = BatchTracker(num_init_batches=1)
        for name in self.data:
            if name in self.extendable_names:
                item = self.data[name]
                if isinstance(item, SparseList):
                    self.data[name] = SparseList(default=item.default)
                else:
                    self.data[name] = []
            else:
                self.data[name] = None

    @property
    def num_iters(self):
        return self.batch_tracker.num_iters

    def add_new(
        self,
        name: str,
        extendable: bool,
        sparse: bool = False,
        default: Optional[Any] = None,
    ) -> None:
        """Adds a new result of the specified type.

        Parameters
        ----------
        name : str
            The raw results key in the data dict.
        extendable : bool
            Whether the result is appended to direclty by the BOResults object.
        sparse : bool
            Whether the result is stored in a SparseList.
        default : Optional[Any]
            The default value for sparse results.
        """
        if extendable and not sparse:
            self.data[name] = []
            self.extendable_names.append(name)
        elif extendable and sparse:
            self.data[name] = SparseList(default=default)
            self.extendable_names.append(name)
        else:
            self.data[name] = None

    def update(self, data_new: dict) -> None:
        X = data_new.get("X")
        if X is not None:
            self.batch_tracker.update(len(X))

        for name in data_new.keys() & self.extendable_names:
            item = self.data[name]
            if isinstance(item, SparseList):
                item[self.num_iters - 1] = data_new[name]
            else:
                item.append(data_new[name])
        for name in data_new.keys() - self.extendable_names:
            self.data[name] = data_new[name]

    def set_num_init_batches(self, num_init_batches):
        n_new = num_init_batches
        n_old = self.batch_tracker.num_init_batches
        for name in self.extendable_names:
            subres = self.data[name]
            if isinstance(subres, SparseList):
                data_new = {}
                for i_old, val in subres.items():
                    i_new = i_old + n_old - n_new
                    if i_new >= 0:
                        data_new[i_new] = val
                subres_new = SparseList(data_new, default=subres.default)
                self[name] = subres_new
        self.batch_tracker.num_init_batches = n_new

    def select(self, name: str, itr=None):
        """Provides access to any result from any iteration(s).

        Results are returned batchwise, i.e., if a batch of size B was
        acquired during iteration N, selecting 'X' from itr=N
        will return an array with B rows. 

        Iterations are indexed starting from 0 = all initial data points.
        Note that the iteration number wraps around negative values like
        an numpy array index, s.t. -1 refers to the last iteration.

        Parameters
        ----------
        name : str
            The name of the result. Default results include X, Y, X_next,
            mu_glmin (predicted global min), x_glmin (predicted min location),
            and nu_glmin (predicted min variance).
        itr : int | ArrayLike
            The iteration(s) from which the given result is taken.

        Returns
        -------
        float | np.array
            The result value(s) from the specified iteration(s).
            If more than one iteration was given, the values are
            collected in an array.
        """
        val = self.data[name]
        if itr is not None:
            itr = wrap_index_array(itr, self.num_iters)
            if name in ["X", "Y"]:
                inds = self.batch_tracker.get_multibatch_indices(itr)
                val = val[inds]
            else:
                if np.ndim(itr) == 0:
                    val = val[itr]
                else:
                    val = np.array([val[i] for i in itr])
        return val

    @functools.lru_cache(1)
    def reconstruct_acq_func(self, itr: int) -> BaseAcquisition:
        """Recontructs the acquisition function for a given iteration.

        The last call is cached to prevent fitting the 
        same model over and over again for repeated calls.

        Parameters
        ----------
        itr : int
            The target iteration.

        Returns
        -------
        acqfn: BaseAcquisition
            The user specified acquisition function 
            with model data taken from iterations 0,...,itr.
        """
        itr = wrap_index(itr, self.num_iters)
        model = self.reconstruct_model(itr)
        acqfn = copy.deepcopy(self.settings.acqfn)
        acqfn.model = model

        # Return Explore acq func if pure exploration
        # was triggered this iteration.
        # acqtol = self.settings['acqtol']
        # if acqtol is not None:
        #     x_next = self.get_next_acq(itr)
        #     var_next = model.predict(x_next)[1]
        #     if var_next < acqtol**2:
        #         acqfn = Explore(model)

        return acqfn

    @functools.lru_cache(1)
    def reconstruct_model(self, itr: int) -> Model:
        """Recontructs the GPR model for a given iteration.

        The last call is cached to prevent fitting the 
        same model over and over again for repeated calls.

        Parameters
        ----------
        itr : int
            The target iteration.

        Returns
        -------
        model: Model
            A BOSS model instance with data and hyperparameters
            taken from iterations 0,...,itr.
            
        """
        itr = wrap_index(itr, self.num_iters)
        X = self.select("X", np.arange(itr + 1))
        Y = self.select("Y", np.arange(itr + 1))
        kernel = KernelFactory.construct_kernel(self.settings)
        noise = self.settings["noise"]
        ynorm = self.settings["ynorm"]
        model = Model(X, Y, kernel, noise, ynorm)
        params = self.select("model_params", itr)
        if params is not None:
            model.set_unfixed_params(params)
        else:
            model.optimize()
            self['model_params'][itr] = model.get_unfixed_params()
        return model

    def calc_missing_minima(self) -> None:
        """Calculates global minima for the given iteration(s).

        If a minima if found to be missing from the results, it 
        will be stored (in addition to being returned from the function).
        """
        for itr in range(self.num_iters):
            if self['mu_glmin'].is_default(itr):
                model = self.reconstruct_model(itr)
                x_glmin, mu_glmin, nu_glmin = minimize_model(
                    model,
                    self.settings["bounds"],
                    self.settings['optimtype'],
                    self.settings["kernel"],
                    self.settings["min_dist_acqs"],
                    accuracy=self.settings["minzacc"],
                )
                self['x_glmin'][itr] = x_glmin
                self['mu_glmin'][itr] = mu_glmin
                self['nu_glmin'][itr] = nu_glmin

    def get_best_acq(self, itr_max: int = -1) -> Tuple[np.ndarray, np.ndarray]:
        """Finds the best acquisition = the one with the lowest y-value.

        Parameters
        ----------
        itr_max : int
            Restricts the acquisitions included in the search to
            0, ..., itr_max.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            The best acquisition as a tuple (x, y).
            
        """
        itr_max = wrap_index(itr_max, self.num_iters)
        X = self.select("X", np.arange(itr_max + 1))
        Y = self.select("Y", np.arange(itr_max + 1))
        i_best = np.argmin(Y[:,0])
        x_best = X[i_best]
        y_best = Y[i_best]
        return x_best, y_best

    def get_est_yrange(self, itr_max: int = -1) -> np.ndarray:
        """Finds the estimated range of y.

        Parameters
        ----------
        itr_max : int
            Restricts the acquisitions included in the search to
            0, ..., itr_max.

        Returns
        -------
        tuple
            The estimated y-range as a tuple (y_low, y_high).
        """
        itr_max = wrap_index(itr_max, self.num_iters)
        Y = self.select("Y", np.arange(itr_max + 1))
        return np.min(Y), np.max(Y)

    def get_next_acq(self, itr: int = -1) -> np.ndarray:
        itr = wrap_index(itr, self.num_iters)
        if itr == self.batch_tracker.num_iters - 1:
            return self['X_next']
        else:
            return self.select('X', itr + 1)

    @classmethod
    def from_file(cls, rstfile, outfile):
        """Restores Results from boss output files.

        Parameters
        ----------
        rstfile : Union[str, Path]
            A BOSS restart file (boss.rst)
        outfile : Union[str, Path]
            A BOSS output file (boss.out)

        Returns
        -------
        BOResults
            A BOResults object pre-filled with data from the output files.
        """
        self = cls.__new__(cls)
        self.data = {}
        self.extendable_names = ['mu_glmin', 'nu_glmin', 'x_glmin', 'model_params']
        settings = Settings.from_file(rstfile)
        self.settings = settings

        # Assign batch data.
        dim = settings.dim
        acqs, batch_ind_lims = parse.parse_acqs(
            settings,
            outfile,
            self.settings['ygrad'],
        )
        self.data['X'] = acqs[:, :dim]
        self.data['Y'] = acqs[:, dim:]

        self.batch_tracker = BatchTracker(settings['initpts'])
        self.batch_tracker._ind_lims = batch_ind_lims

        # lambda for converting ensemble size to iteration number
        to_itr = lambda n: np.searchsorted(self.batch_tracker.ensemble_sizes, n)

        # Assign global min data.
        data_min = parse.parse_min_preds(settings, outfile)
        npts = data_min[:, 0].astype(int)
        x_glmin = data_min[:, 1: 1 + dim]
        mu_glmin = data_min[:, -2]
        nu_glmin = data_min[:, -1]
        self.data['x_glmin'] = SparseList(
            {to_itr(n): x for n, x in zip(npts, x_glmin)}
        )
        self.data['mu_glmin'] = SparseList(
            {to_itr(n): mu for n, mu in zip(npts, mu_glmin)}, default=np.nan
        )
        self.data['nu_glmin'] = SparseList(
            {to_itr(n): nu for n, nu in zip(npts, nu_glmin)}, default=np.nan
        )

        # Assign model params.
        data_params = parse.parse_mod_params(outfile)
        npts = data_min[:, 0].astype(int)
        self.data['model_params'] = SparseList(
            {to_itr(n): x for n, x in zip(npts, data_params[:, 1:])}
        )

        # Assign next acquisitions.
        data_next = parse.parse_xnexts(outfile)
        self.data['X_next'] = data_next[-1, 1:]
        return self



import datetime
import os
from contextlib import contextmanager

import numpy as np

import boss.io.ioutils as io
import boss.keywords as bkw
from boss import __version__
from boss.utils.timer import Timer


class MainOutput:
    """
    Functionality to write to the main output (*.out) file.
    """

    def __init__(self, settings, timer=None):
        self.ipfile = settings["ipfile"]
        self.outfile = settings["outfile"]
        self.settings = settings
        self.ensemble_size = 0
        self.curr_pts = 1
        if timer is None:
            self.timer = Timer()
        else:
            self.timer = timer

    def new_file(self):
        """Intializes a new main output file with header, input file and settings."""
        if os.path.isfile(self.outfile):
            print("warning: overwriting file '" + self.outfile + "'")

        self.header()
        if os.path.isfile(self.ipfile):
            self.ipfile_repeat(self.settings)

        self.add_settings(self.settings)

    def header(self):
        """
        Writes a header to main output file overwriting a possibly existing
        old output file at the same filepath.
        """
        s = (
            "\n-----------------------------   Welcome to ....   ----------"
            + "--------------------\n"
            + "                      _______  _______ _______ _______ \n"
            + "                     |   _   \|   _   |   _   |   _   |\n"
            + "                     |.  1   /|.  |   |   1___|   1___|\n"
            + "                     |.  _   \|.  |   |____   |____   |\n"
            + "                     |:  1    |:  1   |:  1   |:  1   |\n"
            + "                     |::.. .  |::.. . |::.. . |::.. . |\n"
            + "                     `-------'`-------`-------`-------'\n\n"
            + "{:^80s}\n".format("Version " + str(__version__))
            + "{:^80s}\n".format(datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S"))
            + "------------------------------------------------------------"
            + "--------------------\n\n"
        )
        with open(self.outfile, "w") as fd:
            fd.write(s)

    def footer(self, totaltime):
        """
        Writes a footer to main output file
        """
        text = "BOSS is done! Have a nice day :)"
        s = (
            "\n\n--------------------------------------------------------"
            + "------------------------\n"
        )
        s += "{:^80s}\n\n".format(text)
        s += "{:^40s}".format(
            datetime.datetime.now().strftime("Datetime %d-%m-%Y %H:%M:%S")
        )
        s += "{:^40s}".format("Total time [s] %s" % (totaltime))  # code total time
        s += (
            "\n---------------------------------------------------------"
            + "-----------------------\n"
        )
        with open(self.outfile, "a") as fd:
            fd.write(s)

    def ipfile_repeat(self, is_rst):
        """
        Repeats the input file near the beggining of the main output file.
        """
        self.progress_msg("Reading BOSS input file from: " + self.ipfile)
        self.progress_msg("Initializing...\n")
        self.section_header("INPUT FILE")
        s = ""
        with open(self.ipfile, "r") as f:
            line = f.readline()
            while len(line) > 0:
                s += line
                line = f.readline()

        with open(self.outfile, "a") as fd:
            fd.write(s)

    def add_settings(self, settings):
        """
        Outputs the interpreted code variable settings to main output file.
        """
        self.section_header("SIMULATION OPTIONS")
        s = "|| File input/output \n"
        s += "ipfile         %s\n" % (settings["ipfile"])
        s += "userfn         %s\n" % (bkw.stringify(settings["userfn"]))
        s += "outfile        %s\n" % (settings["outfile"])
        s += "rstfile        %s\n" % (settings["rstfile"])

        s += "\n|| Key settings \n"
        s += (
            "bounds        "
            + io.twoDfloatarray_line(settings["bounds"], settings.dim, 2)
            + "\n"
        )
        s += (
            "kerntype      "
            + io.oneDarray_line(settings["kernel"], settings.dim, str)
            + "\n"
        )
        if np.any(settings["kernel"] == "stdp"):
            s += (
                "periods      "
                + io.oneDarray_line(settings["periods"], settings.dim, float)
                + "\n"
            )
        s += (
            "yrange        "
            + io.oneDarray_line(settings["yrange"], 2, float)
            + "\n"
        )
        s += "noise          %8.3E\n" % (settings["noise"])

        if not settings.is_rst:
            s += "inittype       %s\n" % (settings["inittype"])
        s += "initpts   %i    iterpts   %i\n" % (
            settings["initpts"],
            settings["iterpts"],
        )

        if settings["glmin_tol"] is not None:
            s += "glmin_tol         %8.3E   %i\n" % (
                settings["glmin_tol"][0],
                settings["glmin_tol"][1],
            )
        else:
            s += "glmin_tol         none\n"

        s += "\n|| Data acquisition \n"
        s += "acqfn                %s\n" % (settings["acqfn_name"])
        s += "acqtol               "
        if settings["acqtol"] is None:
            s += "none\n"
        else:
            s += "%8.3E\n" % (settings["acqtol"])

        s += "\n|| GP hyperparameters \n"
        s += (
            "thetainit      "
            + io.oneDarray_line(settings["thetainit"], settings.dim, float)
            + "\n"
        )
        s += (
            "thetabounds    "
            + io.twoDfloatarray_line(settings["thetabounds"], settings.dim + 1, 2)
            + "\n"
        )
        s += "thetaprior      %s\n" % (settings["thetaprior"])
        s += (
            "thetapriorpar  "
            + io.twoDfloatarray_line(settings["thetapriorpar"], settings.dim + 1, 2)
            + "\n"
        )

        s += "\n|| Hyperparameter optimization\n"
        s += "optimtype      %s\n" % settings['optimtype']
        s += "updatefreq   %i  initupdate      %s\n" % (
            settings["updatefreq"],
            str(settings["initupdate"]),
        )
        s += "updateoffset %i  updaterestarts  %i\n\n" % (
            settings["updateoffset"],
            settings["updaterestarts"],
        )
        s += "\n|| Miscellaneous\n"
        if settings['seed'] is not None:
            s += "seed      %i\n" % settings['seed']

        s += "\n|| Postprocessing\n"
        s += "pp_iters          = "
        if settings['pp_iters'] is None:
            s += "None (all iterations will be used if postprocessing is called)\n"
        else:
            s += f"[{settings['pp_iters'][0]} ... {settings['pp_iters'][-1]}]\n"
        # s += io.oneDarray_line(settings["pp_iters"],len(settings["pp_iters"]),int)+"\n"
        s += "pp_models         = %s\n" % (settings["pp_models"])
        s += "pp_acq_funcs      = %s\n" % (settings["pp_acq_funcs"])
        s += "pp_truef_npts     = "
        if settings["pp_truef_npts"] is None:
            s += "none\n"
        else:
            s += "%i\n" % (settings["pp_truef_npts"])
        s += "pp_model_slice        ="
        s += (
            io.oneDarray_line(
                settings["pp_model_slice"], len(settings["pp_model_slice"]), int
            )
            + "\n"
        )
        if settings["pp_var_defaults"] is not None:
            s += "pp_var_defaults   ="
            s += (
                io.oneDarray_line(
                    settings["pp_var_defaults"],
                    len(settings["pp_var_defaults"]),
                    float,
                )
                + "\n"
            )
        s += "pp_truef_at_glmins = %s\n" % (settings["pp_truef_at_glmins"])
        s += "pp_local_minima   = "
        if settings["pp_local_minima"] is None:
            s += "none\n\n"
        else:
            s += "%i\n\n" % (settings["pp_local_minima"])

        with open(self.outfile, "a") as fd:
            fd.write(s)

    def progress_msg(self, msg, preceding_bl=False, nospace=False):
        """
        Announce progress message to main output file depending on verbosity.
        """
        m = ""
        if preceding_bl:
            m = m + "\n"
        m = m + "|"
        if not nospace:
            m = m + " "
        m = m + msg + "\n"

        with open(self.outfile, "a") as fd:
            fd.write(m)

    def convergence_stop(self):
        """
        Announces BO stop due to global minimum convergence
        """
        msg = "Stopped BO due to global minimum prediction convergence"
        self.progress_msg(msg, 0, True)

    def initpt_summary(self, X_new, Y_new, timer):
        s = "| Data point added to dataset (x y): \n"
        for i in range(len(Y_new)):
            s += io.data_line(X_new[i], [Y_new[i]], fstr="%18.10E")
        self.ensemble_size += len(X_new)
        s += f"\n| Total ensemble size: {self.ensemble_size}\n"
        s += f"\nIteration time [s]: {timer.getLapTime():8.3f}"
        s += f"        Total time [s]: {timer.getTotalTime():8.3f}"
        with open(self.outfile, "a") as fd:
            fd.write(s)

    def section_header(self, name, number=None):
        if number:
            text = f"{name} {number}"
        else:
            text = name
        s = (
            "\n--------------------------------------------------------"
            + "------------------------\n"
        )
        s += "{:^80}".format(text)
        s += (
            "\n---------------------------------------------------------"
            + "-----------------------\n"
        )
        with open(self.outfile, "a") as fd:
            fd.write(s)

    def iteration_summary(
        self,
        results,
        timer,
    ):
        """
        Outputs info about one BO iteration to main output file
        """
        res = results

        s = "| Data point added to dataset (x y): \n"
        X_new = res.select("X", -1)
        Y_new = res.select("Y", -1)
        ensemble_size = res.batch_tracker.ensemble_sizes[-1]

        for i in range(len(Y_new)):
            s += io.data_line(X_new[i], Y_new[i], fstr="%18.10E")
        s += f"\n| Total ensemble size: {ensemble_size}\n"

        # Print certain results only if initialization is done.
        if self.curr_pts >= self.settings['initpts']:

            # Best acquisitions
            x_best, y_best = res.get_best_acq()
            s += "| Best acquisition, x_best y_best:\n"
            s += io.data_line(x_best.flatten(), y_best.flatten(), fstr="%18.10E")

            # Convergence stuff
            mu_glmins = res["mu_glmin"]
            x_glmins = res["x_glmin"]
            nu_glmins = res["nu_glmin"]
            s += "| Global minimum prediction, x_glmin mu_glmin +- nu_glmin:\n"
            s += io.data_line(
                x_glmins.value(-1).flatten(),
                [mu_glmins.value(-1), nu_glmins.value(-1)],
                fstr="%18.10E",
            )
            if len(mu_glmins.data) > 1:
                dx_glmin = np.linalg.norm(x_glmins.value(-1) - x_glmins.value(-2))
                y_range = res.get_est_yrange()
                dmu_glmin = (
                    abs(mu_glmins.value(-1) - mu_glmins.value(-2)) /
                    (y_range[1] - y_range[0]) 
                )
                s += "| Global minimum convergence, dx_glmin dmu_glmin:\n"
                s += io.data_line([dx_glmin, dmu_glmin], fstr="%18.10E")

            s += "\n| GP model hyperparameters:\n"
            mod_par_last = res['model_params'].value(-1)
            if mod_par_last is not None:
                s += io.data_line(mod_par_last, fstr="%18.10E")

            s += "| Next sampling location x_next:\n"
            s += io.data_line(res['X_next'], fstr="%18.10E")

        s += f"\nIteration time [s]: {timer.getLapTime():8.3f}"
        s += f"        Total time [s]: {timer.getTotalTime():8.3f}"

        with open(self.outfile, "a") as fd:
            fd.write(s + "\n")

    @contextmanager
    def summarize_results(self, results):
        initpts = self.settings['initpts']
        self.timer.startLap()
        if self.curr_pts <= initpts:
            self.section_header("INITIAL DATAPOINT", self.curr_pts)
        else:
            self.section_header("BO ITERATION", self.curr_pts - initpts)
        yield
        self.iteration_summary(results, self.timer)
        self.curr_pts += 1

    def mep_start(self, mep):
        """
        Writes MEP options and local minima
        """
        s = ""
        s += "|| MEP options\n"
        s += "precision %8d    maxE      %8.3E\n" % (mep.precision, mep.maxE)
        s += "rrtsteps  %8d    nebsteps   %8d\n" % (mep.rrtsteps, mep.nebsteps)
        s += "\n"
        s += "Energy threshold starting at %8.3E, stepsize %8.3E.\n" % (
            mep.e_start,
            mep.stepsize,
        )
        s += "\n"

        s += "|| Minima\n"
        s += "(pt index, coordinates)\n"
        for i in range(mep.min_points.shape[0]):
            s += str(i)
            for j in range(mep.min_points.shape[1]):
                s += " " + str(mep.min_points[i, j])
            s += "\n"
        s += "\n"

        with open(self.outfile, "a") as fd:
            fd.write(s)

    def mep_result(self, mep):
        """
        Writes the results of MEP
        """
        s = "\n"
        s += "|| MEP results\n"
        s += "(pt index, pt index, highest energy on the minimum energy path)\n"
        l = mep.min_points.shape[0]
        e = np.zeros((l, l))
        for path in mep.fullpaths:
            e[path.mi, path.mj] = path.maxe

        for i in range(l):
            for j in range(l):
                if i < j:
                    s += "%d %d %8.3E\n" % (i, j, e[i, j])

        s += "\n"
        with open(self.outfile, "a") as fd:
            fd.write(s)



"""A collection of array-related functions.
"""
import numpy as np


def shape_consistent_X(X, dim):
    """Ensures that X-data is consistent with the user function dimension.

    The shape of X is checked for consistency with the user function dimension,
    (as specified by the bounds). 

    Parameters
    ----------
    X : np.ndarray
        N x dim array of x-data. 1d-arrays will be promoted to 2d.
    dim : int
        Dimension of the user function domain, as specified by the bounds.

    Returns
    ------
    np.ndarray
        X-data with consistent shape.
    """
    X = np.atleast_2d(X)
    if dim == 1:
        if X.shape[1] > dim:
            X = X.T
    else:
        if X.shape[1] != dim:
            raise ValueError(f'X-shape = {X.shape} inconsistent with dimension of bounds = {dim}')
    return X


def shape_consistent_Y(Y, dim, ygrad=False):
    Y = np.atleast_2d(Y)
    if not ygrad:
        if Y.shape[1] > 1:
            Y = Y.T
    else:
        if Y.shape[1] != (dim + 1):
            if (Y.shape[0] != (dim + 1)):
                raise ValueError('Missing gradient data in Y.')
            else:
                Y = Y.T
    return Y


def shape_consistent_XY(X, Y, dim, nan_pad=False, ygrad=False):
    """Ensures that X and Y-data are shape consistent.

    The shape of X is checked for consistency with the user function dimension,
    (as specified by the bounds). The shape of Y is then checked for consistency
    with X and possibly padded with nan-values upon request.

    This function should be called by any user-facing method that accepts X, Y-data.

    Parameters
    ----------
    X : np.ndarray
        N x dim array of x-data. 1d-arrays will be promoted to 2d.
    Y : np.ndarray
        N x 1 array of y-data. 1d-arrays will be promoted to 2d.
        if ygrad:
            N x (1 + dim) array of y- and dy-data.
    dim : int
        Dimension of the user function domain, as specified by the bounds.
    nan_pad : bool = False
        Whether to allow Y with less rows than X to be nan-padded until
        the number of rows match.
    ygrad : bool = False
        Whether gradient observations are included or not.

    Returns
    ------
    tuple
        X, Y-data with consistent shapes.
    """
    X = shape_consistent_X(X, dim)
    if Y is None:
        if nan_pad:
            Y = np.empty((X.shape[0], 1 + ygrad * dim)) * np.nan
        else:
            raise ValueError('Y=None not allowed for nan_pad=False')
    else:
        Y = shape_consistent_Y(Y, dim, ygrad)
        n_diff = X.shape[0] - Y.shape[0]
        if n_diff > 0:
            if nan_pad:
                Y_fill = np.empty((n_diff, 1 + ygrad * dim)) * np.nan
                Y = np.concatenate((Y, Y_fill), axis=0)
            else:
                raise ValueError('Number of rows in X and Y must match.')
        elif n_diff < 0:
            raise ValueError('Y cannot contain more rows than X.')
    return X, Y



"""
The format of the input file is a set of space-separated keyword value pairs:

    # comment
    key1 val1
    key2 val2  # comment2

Keyword values are evaluated according to the ruleset:

    foo               -> str
    3                 -> int
    1.0               -> float
    foo bar           -> List[str]
    3 10              -> np.ndarray(shape(1, 2), dtype=int)
    5.0 7.0           -> np.ndarray(shape(1, 2), dtype=float)
    1.0 2.0 ; 3.0 4.0 -> np.ndarray(shape=(2,2), dtype=float)
"""
from pathlib import Path
import boss.keywords as bkw

import numpy as np


def parse_input_file(file_path, skip=None):
    """Parses a boss input file.

    Parses the contents of the file and organizes keywords and their
    values into a dictionary. If present, restart data present under
    the optional results header is also parsed.

    Parameters
    ----------
    input_file : Optional[Union[str, Path]]
        Input file path.

    skip: Optional[str] = None
        Skip parsing part of the input, can be either 'keywords' or 'results'.

    Returns
    -------
    input_data: dict
        A dictionary containing the parsed input data. Can contain
        the following items:

        is_rst : bool
            Whether rst data was present in the input_file or not

        rst_data: np.ndarray
            Array containing rst data, each row is composed of the X, y data
            followed by the model parameters. Rows containing the initial
            points and/or missing y-data are padded with np.nan.
    """
    input_data = {"is_rst": False}
    with open(file_path, "r") as fd:
        text = fd.read()

    # The strategy is to split the text around the results header
    # and then parse the two parts seprately.
    text_split = text.split("RESULTS:")

    # The first part of the split contains all boss keywords.
    input_data["keywords"] = {}
    if skip != "keywords":
        keywords = input_data["keywords"]
        kw_text = text_split[0].strip()
        kw_lines = kw_text.splitlines()
        for ln in kw_lines:
            ln = ln.split("#", maxsplit=1)[0].rstrip()
            if len(ln) > 0:
                key, val_str = ln.split(maxsplit=1)
                category = bkw.find_category(key)
                keywords[key] = bkw.destringify(val_str, category)

    # The second part of the split contains the restart data.
    if len(text_split) == 2:
        input_data["is_rst"] = True
        if skip != "results":
            rst_text = text_split[1].strip()
            rst_lines = rst_text.splitlines()

            # If no actual rst data is present we return early.
            if len(rst_lines) == 0:
                return input_data

            arrs = []
            for ln in rst_lines:
                ln = ln.split("#", maxsplit=1)[0].rstrip()
                if len(ln) > 0:
                    arrs.append(np.fromstring(ln, dtype=float, sep=" "))

            totalpts = len(arrs)
            max_len = max([len(a) for a in arrs])
            rst_data = np.full((totalpts, max_len), np.nan)
            for i, arr in enumerate(arrs):
                rst_data[i, : len(arr)] = arr

            input_data["rst_data"] = rst_data
    return input_data


def parse_min_preds(settings, outfile):
    """
    Extracts x_glmin, mu_glmin and nu_glmin from all bo iterations in the output
    file. Returns a numpy array where each row is [npts, x_glmin, mu_glmin,
    nu_glmin]. Works on output files with all verbosity levels.
    """
    data = []
    npts = None

    with open(outfile, "r") as f:
        line = f.readline()
        while len(line) > 0:
            if line.find("Total ensemble size") != -1:
                npts = int(line.split()[-1])
            elif line.find("Global minimum prediction") != -1:
                line = f.readline()
                line = line.split()
                data.append(np.concatenate([[npts], line[: settings.dim + 2]]))
            line = f.readline()

    return np.array(data).astype(float)


def parse_acqs(settings, outfile, ygrad=False):
    acqs = []
    batch_ind_limits = [0]
    n_values = 1 + settings.dim * (1 + ygrad)
    with open(outfile, "r") as f:
        line = f.readline()
        while len(line) > 0:
            if line.find("Data point added to dataset") != -1:
                line = f.readline()
                ind_lim = 0
                while len(line.split()) == n_values and line.find(".") != -1:
                    ind_lim += 1
                    acq = np.fromstring(line, sep=' ', dtype=float)
                    acqs.append(acq)
                    line = f.readline()
                batch_ind_limits.append(batch_ind_limits[-1] + ind_lim)
            else:
                line = f.readline()
    return np.array(acqs), batch_ind_limits


def parse_best_acqs(settings, outfile):
    """
    Extracts xbest and ybest from all bo iterations in the output file.
    Returns a numpy array where each element is [npts, xbest, ybest].
    Works only on output files with verbosity levels of at least 1.
    """
    data = []
    npts = None

    with open(outfile, "r") as f:
        line = f.readline()
        while len(line) > 0:
            if line.find("Total ensemble size") != -1:
                npts = int(line.split()[-1])
            elif line.find("Best acquisition") != -1:
                line = f.readline()
                line = line.split()
                data.append(np.concatenate([[npts], line[: settings.dim + 1]]))
            line = f.readline()

    return np.array(data).astype(float)


def parse_conv_measures(settings, outfile):
    """
    Extracts dx_glmin and dmu_glmin from all bo iterations in the output
    file. Returns a numpy array where each element is
    [npts, dx_glmin, dmu_glmin].
    Works only on output files with verbosity levels of at least 1.
    """
    data = []
    npts = None

    with open(outfile, "r") as f:
        line = f.readline()
        while len(line) > 0:
            if line.find("Total ensemble size") != -1:
                npts = int(line.split()[-1])
            elif line.find("Global minimum convergence") != -1:
                line = f.readline()
                line = line.split()
                if "none" in line:
                    pass
                else:
                    data.append(np.concatenate([[npts], line]))
            line = f.readline()

    return np.array(data).astype(float)


def parse_mod_params(outfile):
    """
    Extracts unfixed GP model hyperparameters from all bo iterations in the
    output file. Returns a numpy array where each element is
    [npts, unfixed model parameters].
    Works only on output files with verbosity levels of at least 2.
    """
    hpar = []
    npts = None

    with open(outfile, "r") as f:
        line = f.readline()
        while len(line) > 0:
            if line.find("Total ensemble size") != -1:
                npts = int(line.split()[-1])
            elif line.find("GP model hyperparameters") != -1:
                line = f.readline()
                line = line.split()
                hpar.append(np.concatenate([[npts], line]))
            line = f.readline()

    return np.array(hpar).astype(float)


def parse_xnexts(outfile):
    """
    Extracts xnext locations from all bo iterations in the output file.
    Returns a numpy array where each element is [npts, xnext].
    Note that this means that the location 'xnext' is to be evaluated
    in the next iteration 'iter+1'.
    Works only on output files with verbosity levels of at least 2.
    """
    try:
        data = []
        with open(outfile) as f:
            data = []
            line = f.readline()
            npts = None
            while len(line) > 0:
                if line.find("Total ensemble size") != -1:
                    npts = int(line.split()[-1])
                elif line.find("Next sampling location") != -1:
                    line = f.readline()
                    line = line.split()
                    data.append(np.concatenate([[npts], line]))
                line = f.readline()
            return np.array(data).astype(float)
    except:
        raise OSError("Error trying to read file '" + outfile + "'")


def parse_minima(outfile):
    """
    Extract local minima
    """
    try:
        f = open(outfile)
        line = f.readline()
        if line[:14] == "# Local minima":
            ok = True
        else:
            ok = False
        f.close()
    except:
        raise OSError("Error trying to read file '" + outfile + "'")

    if not ok:
        raise Exception("'" + outfile + "' not recognized as " + "a minima file.")

    try:
        f = open(outfile)
        line = f.readline()
        line = f.readline()
        minima = np.array(line.split())
        line = f.readline()
        while len(line) > 0:
            minima = np.vstack((minima, line.split()))
            line = f.readline()
        f.close()
        return minima.astype(float)

    except:
        raise OSError("Error trying to read file '" + outfile + "'")



import numpy as np
import scipy.stats as stats

from boss.bo.acq.base import BaseAcquisition


class EI(BaseAcquisition):
    """
    Expected improvement acquisition function.

    Takes one parameter. Parameter values > 0 boost exploration.
    """

    def __init__(self, jitter=0, model=None):
        """
        Parameters
        ----------
        jitter : float, optional
            Parameter > 0 to boost exploration.
        """
        super().__init__(model=model)
        self.jitter = jitter

    def evaluate(self, x):
        """
        Compute the expected improvement at location x.

        Parameters
        ----------
        x : ndarray
          Input to the acquisition function. 2D array containing data with
          'float' type.

        Returns
        -------
        f_acq : ndarray
          Acquisition function evaluated at 'x'. 2D array containing data with
          'float' type.
        """
        best_acq = np.min(self.model.Y)
        mean, var = self.model.predict(np.atleast_2d(x), noise=False)
        sd = np.sqrt(var)
        z = (best_acq - mean - self.jitter) / sd
        phi = stats.norm.pdf(z)
        Phi = stats.norm.cdf(z)
        f_acqu = -sd * (z * Phi + phi)
        return f_acqu

    def evaluate_with_gradient(self, x):
        """
        Compute the expected improvement and its gradient at location x.
        Gradients are required to apply L-BFGS-B minimizer.

        Parameters
        ----------
        x : ndarray
          Input to the acquisition function. 2D array containing data with
          'float' type.

        Returns
        -------
        f_acq, df_acq : tuple of ndarray
          Acquisition function and it's gradient, evaluated at 'x'. Tuple of
          2D arrays, containing data with 'float' type.
        """
        best_acq = np.min(self.model.Y)
        mean, var = self.model.predict(np.atleast_2d(x), noise=False)
        sd = np.sqrt(var)
        dmean, dvar = self.model.predict_grads(np.atleast_2d(x))
        dmean = dmean[:, :, 0]
        z = (best_acq - mean - self.jitter) / sd
        phi = stats.norm.pdf(z)
        Phi = stats.norm.cdf(z)
        f_acqu = -sd * (z * Phi + phi)
        df_acqu = dmean*Phi - dvar/(2*sd) * phi
        df_acqu = df_acqu.squeeze()
        return f_acqu, df_acqu

    @property
    def has_gradient(self):
        """Property to inform whether acquisition function has a gradient.

        Returns
        -------
        bool
        """
        return True



from abc import ABC, abstractmethod
import GPy
import numpy as np


class BaseModel(ABC):
    """
    Base class for surrogate models used in Bayesian optimization.
    """

    # predictions:

    @abstractmethod
    def predict(self, x, noise=True):
        """
        Returns model prediction mean and variance at point x, with or without
        model variance (noise).
        """
        pass

    @abstractmethod
    def predict_grads(self, x):
        """
        Returns prediction mean and variance gradients with respect to input
        at point x.
        """
        pass

    @abstractmethod
    def predict_mean_sd_grads(self, x, noise=True):
        """
        Returns the model prediction mean, standard deviation and their
        gradients at point x, with or without model variance (noise).
        """
        pass

    @abstractmethod
    def predict_mean_grad(self, x):
        """
        Returns model mean and its gradient at point x.
        """
        pass

    @abstractmethod
    def estimate_num_local_minima(self, search_bounds):
        """
        Returns estimated number of local minima within bounds, calculated
        based on model properties.
        """
        pass

    # model parameters:

    @abstractmethod
    def get_all_params(self):
        """
        Returns model parameters as a dictionary.
        """
        pass

    @abstractmethod
    def get_unfixed_params(self):
        """
        Returns the unfixed parameters of the model in an array.
        """
        pass

    @abstractmethod
    def sample_unfixed_params(self, num_samples):
        """
        Sample unfixed model parameters.
        """
        pass

    @abstractmethod
    def set_unfixed_params(self, params):
        """
        Sets the unfixed parameters of the model to given values.
        """
        pass

    @abstractmethod
    def optimize(self):
        """
        Updates unfixed model parameters.
        """
        pass

    # observations:

    @abstractmethod
    def add_data(self, X_new, Y_new):
        """
        Updates the model evidence (observations) dataset appending.
        """
        pass

    @abstractmethod
    def redefine_data(self, X, Y):
        """
        Updates the model evidence (observations) dataset overwriting.
        """
        pass

    @abstractmethod
    def get_best_xy(self):
        """
        Returns the lowest energy acquisition (x, y).
        """
        pass

    @property
    @abstractmethod
    def kernel(self):
        pass

    @property
    @abstractmethod
    def X(self):
        pass

    @property
    @abstractmethod
    def Y(self):
        pass


class Model(BaseModel):
    """
    Functionality for creating, refitting and optimizing a GP model
    """

    def __init__(self, X, Y, kernel, noise, ynorm):
        """
        Initializes the Model class.
        """
        self.dim = kernel.input_dim
        # normalise observation mean:
        self.normmean = np.mean(Y)
        # scale normalisation is not used unless ynorm is true:
        self.use_norm = ynorm
        # previous boss code used normsd to normalise observation variance:
        # if self.ynorm: self.normsd = np.std(Y)
        # current version normalises observation range:
        self.normsd = np.ptp(Y) if self.use_norm else 1
        # note that the choice betweeen variance or range normalisation needs
        # to be taken into account when we set kernel parameter priors
        # normalised data:
        Y_norm = (Y - self.normmean) / self.normsd
        # initialise model
        self.model = GPy.models.GPRegression(
            X,
            Y_norm,
            kernel=kernel,
            noise_var=noise
        )
        self.model.likelihood.fix()

    # predictions:

    def predict(self, x, noise=True):
        """
        Returns model prediction mean and variance at point x, with or without
        model variance (noise).
        """
        m, v = self.model.predict(np.atleast_2d(x), include_likelihood=noise)
        v = np.clip(v, 1e-12, np.inf)
        return m*self.normsd+self.normmean, v*(self.normsd**2)

    def predict_grads(self, x):
        """
        Returns model prediction mean and variance gradients with respect to
        input at point x.
        """
        dmdx, dvdx = self.model.predictive_gradients(np.atleast_2d(x))
        return dmdx*self.normsd, dvdx*(self.normsd**2)

    def predict_mean_sd_grads(self, x, noise=True):
        """
        Returns the model prediction mean, standard deviation and their
        gradients at point x, with or without model variance (noise).

        This method is a wrapper used primarily during calculations
        of acquisition functions and their derivatives.
        """
        m, v = self.predict(np.atleast_2d(x), noise=noise)
        dmdx, dvdx = self.predict_grads(np.atleast_2d(x))
        dmdx = dmdx[:, :, 0]
        dsdx = dvdx / (2 * np.sqrt(v))
        return m, np.sqrt(v), dmdx, dsdx

    def predict_mean_grad(self, x):
        """Returns model mean and its gradient at point x.

        This method is a wrapper used primarily when the 
        mean function is minimized in order to obtain a
        global minimum prediction.
        """
        m, _ = self.predict(np.atleast_2d(x))
        dmdx, _ = self.predict_grads(np.atleast_2d(x))
        return m, dmdx

    def estimate_num_local_minima(self, search_bounds):
        """
        Returns estimated number of local minima within bounds, calculated
        based on model properties.
        """
        # For the ith dimension, the number of local minima along a slice
        # is approximately n(i) = boundlength(i)/(2*lengthscale(i)). Note
        # that periodic kernels operate on normalised distances: distance
        # between inputs that are period(i)/2 apart is 1. To get the total
        # number of minima for all of the search space, multiply together
        # n(i) over all i.
        numpts = 1
        ks = self.model.kern.parameters if self.dim > 1 else [self.model.kern]
        for bounds, kern in zip(search_bounds, ks):
            if hasattr(kern, 'period'):
                bound_distance = (bounds[1]-bounds[0])/float(kern.period)
            else:
                bound_distance = (bounds[1]-bounds[0])/2
            numpts *= max(1, bound_distance/float(kern.lengthscale))
        return int(numpts)

    # model parameters:

    def get_all_params(self):
        """
        Returns model parameters as a dictionary with entries::
        noise, variance, lengthscales, periods
        where the last two are 1D lists. There exists a period only for those
        dimensions which are using a periodic kernel.
        """
        noise = float(self.model.likelihood.variance)
        sigma = float(self.model.kern.param_array[0])
        lss = []
        pers = []
        ks = self.model.kern.parameters if self.dim > 1 else [self.model.kern]
        for kern in ks:
            lss.append(float(kern.lengthscale))
            if hasattr(kern, 'period'):
                pers.append(float(kern.period))

        # the variables are returned in a dict:
        params = {}
        params['noise'] = noise
        params['variance'] = sigma
        params['lengthscales'] = lss
        params['periods'] = pers

        return params

    def get_unfixed_params(self):
        """
        Returns the unfixed parameters of the model in an array.
        """
        return np.array(self.model.unfixed_param_array.copy()).astype(float)

    def sample_unfixed_params(self, num_samples):
        """
        Sample unfixed model parameters.
        """
        hmc = GPy.inference.mcmc.HMC(self.model)
        burnin = hmc.sample(int(num_samples * 0.33))
        return hmc.sample(num_samples)

    def set_unfixed_params(self, params):
        """
        Sets the unfixed parameters of the model to given values.
        """
        self.model[self.model._fixes_] = params
        self.model.parameters_changed()

    def optimize(self, restarts=1):
        """
        Updates the model hyperparameters by maximizing marginal likelihood.
        """
        self.model.optimization_runs = []
        if restarts == 1:
            self.model.optimize()
        else:
            self.model.optimize_restarts(
                num_restarts=restarts, verbose=False, messages=False
            )

    # observations:

    def add_data(self, X_new, Y_new):
        """
        Updates the model evidence (observations) dataset appending.
        """
        # construct new unnormalised dataset
        X = np.vstack([self.X, np.atleast_2d(X_new)])
        Y = np.vstack([self.Y, Y_new])
        # update model
        self.redefine_data(X, Y)

    def redefine_data(self, X, Y):
        """
        Updates the model evidence (observations) dataset overwriting.
        """
        # update normalisation
        self.normmean = np.mean(Y)
        if self.use_norm:
            self.normsd = np.ptp(Y)
        # update model
        Y_norm = (Y - self.normmean) / self.normsd
        self.model.set_XY(np.atleast_2d(X), np.atleast_2d(Y_norm))

    def get_best_xy(self):
        """
        Returns the lowest energy acquisition (x, y).
        """
        xbest = np.array(self.X[np.argmin(self.Y)])
        ybest = np.min(self.Y)
        return xbest, ybest

    @property
    def kernel(self):
        return self.model.kern

    @property
    def X(self):
        return self.model.X

    @property
    def Y(self):
        return self.model.Y*self.normsd+self.normmean

class GradientModel(BaseModel):
    """
    Functionality for creating, refitting and optimizing a GP model with
    gradient observations.

    The GradientModel utilizes the GPy MultioutputGP model class, which allows
    for multiple input and output channels. We can include observed gradient
    data in GPR by defining separate channels for partial derivatives, in
    addition to the main function value channel.

    The DiffKern kernel computes cross-covariances between channels.
    """

    def __init__(self, X, Y_dY, kernel, noise, ynorm):
        """
        Initializes the GradientModel class.
        """
        self.dim = kernel.input_dim

        # input channels
        X_list = [X] * (self.dim + 1)

        # observations
        Y, dY = Y_dY[:, :1], Y_dY[:, 1:]
        # normalization
        self.use_norm = ynorm
        self.normmean = np.mean(Y)
        self.normsd = np.ptp(Y) if self.use_norm else 1
        Y_norm = (Y - self.normmean) / self.normsd
        # output channels
        Y_list = [Y_norm] + [dY[:, d, None] for d in range(self.dim)]

        # the kernel is accompanied with a DiffKern for each partial derivative.
        kernel_list = [kernel]
        kernel_list += [GPy.kern.DiffKern(kernel, d) for d in range(self.dim)]

        # noise is given to the likelihood.
        likelihood = GPy.likelihoods.Gaussian(variance=noise)
        likelihood_list = [likelihood] * (self.dim + 1)

        # initialize model
        self.model = GPy.models.MultioutputGP(
            X_list=X_list,
            Y_list=Y_list,
            kernel_list=kernel_list,
            likelihood_list=likelihood_list
        )
        self.model.likelihood.fix()

    # predictions:

    def predict(self, x, noise=True):
        """
        Returns model prediction mean and variance at point x, with or without
        model variance (noise).
        """
        m, v = self.model.predict([np.atleast_2d(x)], include_likelihood=noise)
        v = np.clip(v, 1e-12, np.inf)
        return m*self.normsd+self.normmean, v*(self.normsd**2)

    def predict_grads(self, x):
        """
        Returns model prediction mean and variance gradients with respect to
        input at point x.
        """
        dmdx, dvdx = self.model.predictive_gradients([np.atleast_2d(x)])
        return (dmdx*self.normsd)[:, :, None], dvdx*(self.normsd**2)

    def predict_mean_sd_grads(self, x, noise=True):
        """
        Returns the model prediction mean, standard deviation and their
        gradients at point x, with or without model variance (noise).

        This method is a wrapper used primarily during calculations
        of acquisition functions and their derivatives.
        """
        m, v = self.predict(np.atleast_2d(x), noise=noise)
        dmdx, dvdx = self.predict_grads(np.atleast_2d(x))
        dmdx = dmdx[:, :, 0]
        dsdx = dvdx / (2 * np.sqrt(v))
        return m, np.sqrt(v), dmdx, dsdx

    def predict_mean_grad(self, x):
        """Returns model mean and its gradient at point x.

        This method is a wrapper used primarily when the
        mean function is minimized in order to obtain a
        global minimum prediction.
        """
        m, v = self.predict(np.atleast_2d(x))
        dmdx, dvdx = self.predict_grads(np.atleast_2d(x))
        scipygradient = np.asmatrix(dmdx).transpose()
        return m, scipygradient

    def estimate_num_local_minima(self, search_bounds):
        """
        Returns estimated number of local minima within bounds, calculated
        based on model properties.
        """
        # For the ith dimension, the number of local minima along a slice
        # is approximately n(i) = boundlength(i)/(2*lengthscale(i)). Note
        # that periodic kernels operate on normalised distances: distance
        # between inputs that are period(i)/2 apart is 1. To get the total
        # number of minima for all of the search space, multiply together
        # n(i) over all i.
        numpts = 1

        # For the GradientModel, the self.model.kern is the
        # MultioutputDerivativeKern. If self.dim > 1, the Prod kernel which
        # contains the individual kernels is located by
        # self.model.kern.parts[0]. If self.dim == 1, the individual kernel is
        # located by self.model.kern.parts.
        if self.dim > 1:
            ks = self.model.kern.parts[0].parts
        else:
            ks = self.model.kern.parts
        for bounds, kern in zip(search_bounds, ks):
            if hasattr(kern, 'period'):
                bound_distance = (bounds[1] - bounds[0]) / float(kern.period)
            else:
                bound_distance = (bounds[1] - bounds[0]) / 2
            numpts *= max(1, bound_distance / float(kern.lengthscale))
        return int(numpts)

    # model parameters:

    def get_all_params(self):
        """
        Returns model parameters as a dictionary with entries::
        noise, variance, lengthscales, periods
        where the last two are 1D lists. There exists a period only for those
        dimensions which are using a periodic kernel.
        """
        # The MultioutputGP model can contain multiple likelihoods
        # We only use one, and access the noise through model.likelihood[0]
        noise = self.model.likelihood[0]
        sigma = float(self.model.kern.param_array[0])
        lss = []
        pers = []
        # For the GradientModel, the self.model.kern is the
        # MultioutputDerivativeKern. If self.dim > 1, the Prod kernel which
        # contains the individual kernels is located by
        # self.model.kern.parts[0]. If self.dim == 1, the individual kernel is
        # located by self.model.kern.parts.
        if self.dim > 1:
            ks = self.model.kern.parts[0].parts
        else:
            ks = self.model.kern.parts
        for kern in ks:
            lss.append(float(kern.lengthscale))
            if hasattr(kern, 'period'):
                pers.append(float(kern.period))

        # the variables are returned in a dict:
        params = {}
        params['noise'] = noise
        params['variance'] = sigma
        params['lengthscales'] = lss
        params['periods'] = pers

        return params

    def get_unfixed_params(self):
        """
        Returns the unfixed parameters of the model in an array.
        """
        return np.array(self.model.unfixed_param_array.copy()).astype(float)

    def sample_unfixed_params(self, num_samples):
        """
        Sample unfixed model parameters.
        """
        hmc = GPy.inference.mcmc.HMC(self.model)
        burnin = hmc.sample(int(num_samples * 0.33))
        return hmc.sample(num_samples)

    def set_unfixed_params(self, params):
        """
        Sets the unfixed parameters of the model to given values.
        """
        self.model[self.model._fixes_] = params
        self.model.parameters_changed()

    def optimize(self, restarts=1):
        """
        Updates the model hyperparameters by maximizing marginal likelihood.
        """
        self.model.optimization_runs = []
        if restarts == 1:
            self.model.optimize()
        else:
            self.model.optimize_restarts(
                num_restarts=restarts, verbose=False, messages=False
            )

    # observations:

    def add_data(self, X_new, Y_dY_new):
        """
        Updates the model evidence (observations) dataset appending.
        """
        # construct new unnormalized dataset
        X = np.vstack([self.X, np.atleast_2d(X_new)])
        Y_dY = np.vstack([self.Y, Y_dY_new])
        # update model
        self.redefine_data(X, Y_dY)

    def redefine_data(self, X, Y_dY):
        """
        Updates the model evidence (observations) dataset overwriting.
        """
        Y, dY = Y_dY[:, :1], Y_dY[:, 1:]
        # update normalization
        self.normmean = np.mean(Y)
        if self.use_norm:
            self.normsd = np.ptp(Y)
        # update model
        Y_norm = (Y - self.normmean) / self.normsd
        X_list = [X] * (self.dim + 1)
        Y_list = [Y_norm] + [dY[:, d, None] for d in range(self.dim)]
        self.model.set_XY(X_list, Y_list)

    def get_best_xy(self):
        """
        Returns the lowest energy acquisition (x, y).
        """
        xbest = np.array(self.X[np.argmin(self.Y[:,0])])
        ybest = np.min(self.Y[:,0])
        return xbest, ybest

    @property
    def kernel(self):
        return self.model.kern

    @property
    def X(self):
        X_multioutput = self.model.X[:, :-1]
        output_index = self.model.X[:, -1]

        return X_multioutput[np.where(output_index == 0)[0]]

    @property
    def Y(self):
        Y_multioutput = self.model.Y
        output_index = self.model.X[:, -1]

        Y_norm = Y_multioutput[np.where(output_index == 0)[0]]
        Y = Y_norm * self.normsd + self.normmean

        dY = np.empty((len(Y), self.dim), dtype=float)
        for d in range(self.dim):
            dY[:, d, None] = Y_multioutput[np.where(output_index == d + 1)[0]]
        
        return np.concatenate((Y, dY), axis=1)



"""
Plotting functionality. All plot functions make PNG images to destination
file based on given source array(s) of data.
"""

import numpy as np
import matplotlib.pyplot as plt
from boss.utils.arrays import shape_consistent_X

# plt.switch_backend("agg")


def plot_data_acquisitions(settings, dest_file, acqs, min_preds):
    """
    Plots minimum predictions and uncertainty, acquisition energies
    and locations as a function of iteration.
    """
    itNo_acqs = np.clip(acqs[:, 0], int(min_preds[0, 0]), np.inf)
    x = acqs[:, 2 : settings.dim + 2]
    y = acqs[:, settings.dim + 2]
    itNo_mp = min_preds[:, 0]
    xhat = min_preds[:, 2 : settings.dim + 2]
    muhat = min_preds[:, -2]
    nuhat = min_preds[:, -1]

    plt.subplot(211)
    plt.fill_between(itNo_mp, muhat + nuhat, muhat, color="lightgrey")
    plt.fill_between(itNo_mp, muhat - nuhat, muhat, color="lightgrey")
    plt.plot(itNo_mp, muhat + nuhat, "grey", linewidth=2)
    plt.plot(itNo_mp, muhat - nuhat, "grey", linewidth=2)
    plt.plot(itNo_mp, muhat, "black", linewidth=5, label=r"$\mu_{\mathrm{glmin}}$")
    plt.scatter(
        itNo_acqs,
        y,
        s=200,
        linewidth=4,
        facecolors="none",
        edgecolors="black",
        label=r"$acq_y$",
    )

    plt.xlim(min(itNo_mp), max(itNo_mp))
    plt.xticks(itNo_mp[:: max(1, round(len(itNo_mp) / 10))])
    plt.ylabel(r"$y$ with $\mu_{\mathrm{glmin}}$", size=20)
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)

    plt.subplot(212)
    colors = [
        "blue",
        "red",
        "green",
        "brown",
        "yellow",
        "purple",
        "cyan",
        "magenta",
        "grey",
        "gold",
    ]
    for i in range(x.shape[1]):
        plt.plot(
            itNo_mp,
            xhat[:, i],
            color=colors[i % len(colors)],
            linewidth=5,
            label=r"$x_{\mathrm{glmin}}^%i$" % (i + 1),
        )
        plt.scatter(
            itNo_acqs,
            x[:, i],
            s=200,
            linewidth=4,
            facecolors="none",
            edgecolors=colors[i % len(colors)],
            label=r"$acq_{x%i}$" % (i + 1),
        )

    plt.xlim(min(itNo_mp), max(itNo_mp))
    plt.xticks(itNo_mp[:: max(1, round(len(itNo_mp) / 10))])
    plt.xlabel("Iteration", size=20)
    plt.ylabel(r"$x_i$ with $x_{\mathrm{glmin}}^i$", size=20)
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)

    plt.suptitle("Data acquisitions", size=20)
    plt.gcf().set_size_inches(9, 8)
    plt.tight_layout(rect=[0, 0.03, 1, 0.95])
    plt.savefig(dest_file)
    plt.close()


def plot_conv_measures(settings, dest_file, conv_meas, legends=True):
    """
    Plots quantities related to convergence as a function of iteration.
    """
    itNo = conv_meas[:, 0]
    dxhat = conv_meas[:, 2]
    dmuhat = conv_meas[:, 3]

    plt.subplot(211)
    plt.plot(itNo, dxhat, "k", linewidth=5)
    # plt.xlabel('iteration', size=20)
    plt.ylabel(r"$\Delta x_{\mathrm{glmin}}$", size=24)
    plt.xticks(itNo[:: max(1, round(len(itNo) / 10))])
    plt.grid(True)
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)
    if max(dxhat) > 0:
        plt.yscale("log")

    plt.subplot(212)
    plt.plot(itNo, dmuhat, "k", linewidth=5)
    plt.xlabel("Iteration", size=20)
    plt.ylabel(r"$abs(\Delta \mu_{\mathrm{glmin}}/\Delta y$", size=24)
    plt.xticks(itNo[:: max(1, round(len(itNo) / 10))])
    plt.grid(True)
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)
    if max(dmuhat) > 0:
        plt.yscale("log")

    plt.suptitle("Convergence tracking", size=20)
    plt.gcf().set_size_inches(9, 7)
    plt.tight_layout(rect=[0, 0.03, 1, 0.95])
    plt.savefig(dest_file)
    plt.close()


def plot_hyperparameters(settings, dest_file, hypers, legends=True):
    """
    Plots GP model unfixed hyperparameters (variances and lengthscales)
    as a function of iteration.
    """
    itNo = hypers[:, 0]
    var = hypers[:, 2]
    ls = hypers[:, 3:]

    colors = [
        "blue",
        "red",
        "green",
        "brown",
        "yellow",
        "purple",
        "cyan",
        "magenta",
        "black",
        "gold",
    ]
    markers = ["s", "p", "*", "h", "^", "+", "x", "d", "v", "|"]

    plt.subplot(211)
    plt.plot(itNo, var, color="r", linewidth=5)
    plt.ylabel("Variance", size=20)
    plt.xticks(itNo[:: max(1, round(len(itNo) / 10))])
    plt.yscale("log")
    plt.grid(True)
    plt.gca().tick_params(labelsize=18)
    # plt.gca().ticklabel_format(useOffset=False)

    plt.subplot(212)
    for i in range(settings.dim):
        plt.plot(
            itNo,
            ls[:, i],
            color=colors[i % 10],
            linewidth=5,
            linestyle="-",
            marker=markers[i % 10],
            markersize=12,
            label=r"$\ell_ %i$" % (i + 1),
        )
    plt.xlabel("Iteration", size=20)
    plt.ylabel("Lengthscale", size=20)
    plt.xticks(itNo[:: max(1, round(len(itNo) / 10))])
    plt.yscale("log")
    plt.grid(True)
    plt.gca().tick_params(labelsize=18)
    # plt.gca().ticklabel_format(useOffset=False)
    if legends:
        lgd = plt.legend(loc="upper right", prop={"size": 18})

    plt.suptitle("Hyperparameter values", size=20)
    plt.gcf().set_size_inches(11, 7)
    if legends:  # FOR OLD MATPLOTLIB COMPATIBILITY
        plt.savefig(dest_file, bbox_extra_artists=(lgd,), bbox_inches="tight")
    else:
        plt.savefig(dest_file)
    plt.close()


def plot_truef_hat(settings, dest_file, truef_hats, legends=True):
    """
    Plots true function value at xhat locations
    as a function of iteration.
    """
    itNo = truef_hats[:, 0]
    truehat = truef_hats[:, -2]
    last = truehat[-1]
    best = np.array([np.min(truehat[:i]) for i in range(1, len(truehat) + 1)])
    tfhat_muhat = truef_hats[:, -1]

    plt.subplot(211)
    plt.plot(itNo, abs(truehat - last), "k", linewidth=5)
    plt.plot(itNo, abs(best - last), "k", linestyle="dashed", linewidth=3, label="best")
    plt.grid(True)
    plt.ylabel(r"$f(x_{\mathrm{glmin}})-f(x_{\mathrm{glmin,last}})$", size=24)
    plt.xticks(itNo[:: max(1, round(len(itNo) / 10))])
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)
    if max(abs(truehat - last)) and max(abs(best - last)) > 0:
        plt.yscale("log")
    if legends:
        lgd = plt.legend(loc="upper right", prop={"size": 20})

    plt.subplot(212)
    plt.plot(itNo, abs(tfhat_muhat), "k", linewidth=5)
    plt.xlabel("Iteration", size=20)
    plt.ylabel(r"$f(x_{\mathrm{glmin}})-\mu_{\mathrm{glmin}}$", size=24)
    plt.xticks(itNo[:: max(1, round(len(itNo) / 10))])
    plt.grid(True)
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)
    if max(abs(tfhat_muhat)) > 0:
        plt.yscale("log")

    plt.suptitle("True function at minimum predictions", size=20)
    plt.gcf().set_size_inches(9, 9)
    if legends:  # FOR OLD MATPLOTLIB COMPATIBILITY
        plt.savefig(dest_file, bbox_extra_artists=(lgd,), bbox_inches="tight")
    else:
        plt.savefig(dest_file)
    plt.close()


def plot_model(
    settings,
    dest_file,
    model_data,
    xhat=None,
    acqs=None,
    xnext=None,
    minima=None,
    truef=None,
    incl_uncert=True,
    axis_labels=None,
    legends=True,
    paths=None,
):
    """
    Plots a (max 2D) slice of the model.
    """
    if xnext is not None:
        xnext = shape_consistent_X(xnext, settings.dim)

    coords = model_data[:, : settings.dim]
    mu, nu = model_data[:, -2], model_data[:, -1]
    slice_dim = (
        1 if settings["pp_model_slice"][0] == settings["pp_model_slice"][1] else 2
    )
    if axis_labels is None:
        axis_labels = [
            r"$x_%i$" % (settings["pp_model_slice"][0] + 1),
            r"$x_%i$" % (settings["pp_model_slice"][1] + 1),
        ]

    if slice_dim == 1:
        x1 = coords[:, settings["pp_model_slice"][0]]

        if truef is not None:
            crds = truef[:, : settings.dim]
            tf = truef[:, -1]
            tfx1 = crds[:, settings["pp_model_slice"][0]]
            plt.plot(tfx1, tf, "k", linewidth=5, label="f(x)")

        if incl_uncert:
            plt.fill_between(x1, mu + nu, mu, color="lightgrey")
            plt.fill_between(x1, mu - nu, mu, color="lightgrey")
            plt.plot(x1, mu + nu, "grey", linewidth=3, label=r"$\nu(x)$")
            plt.plot(x1, mu - nu, "grey", linewidth=3)

        plt.plot(x1, mu, "b", linewidth=5, label=r"$\mu(x)$")

        if xhat is not None:
            plt.axvline(
                xhat[settings["pp_model_slice"][0]],
                color="red",
                linewidth=5,
                label=r"$x_{\mathrm{glmin}}$",
                zorder=19,
            )

        if settings.dim == 1:
            # plot acqs
            if acqs is not None:
                x1 = acqs[:, settings["pp_model_slice"][0]]
                y = acqs[:, settings.dim]
                plt.scatter(
                    x1,
                    y,
                    s=200,
                    linewidth=6,
                    facecolors="none",
                    edgecolors="brown",
                    label="acqs",
                    zorder=18,
                )

            # plot xnext
            if xnext is not None:
                label = r"$x_{\mathrm{next}}$"
                for xn in xnext:
                    plt.axvline(
                        xn[settings["pp_model_slice"][0]],
                        color="green",
                        linewidth=5,
                        label=label,
                        linestyle="dashed",
                        zorder=20,
                    )
                    label = None

        if settings['pp_var_defaults'] is None and minima is not None:
            x1 = minima[:, settings["pp_model_slice"][0]]
            y = minima[:, -2]
            plt.scatter(
                x1,
                y,
                s=250,
                linewidth=6,
                zorder=19,
                facecolors="none",
                edgecolors="lawngreen",
                label="minima",
            )

        plt.xlim(
            min(coords[:, settings["pp_model_slice"][0]]),
            max(coords[:, settings["pp_model_slice"][0]]),
        )
        yd = max(mu) - min(mu)
        plt.ylim(min(mu) - 0.1 * yd, max(mu) + 0.1 * yd)
        plt.xlabel(axis_labels[0], size=24)
        plt.ylabel(r"$y$", size=24)
        if legends:
            lgd = plt.legend(
                bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
                loc=3,
                ncol=4,
                mode="expand",
                borderaxespad=0.0,
                prop={"size": 20},
            )
        plt.gcf().set_size_inches(10, 8)
        plt.gca().tick_params(labelsize=18)
        plt.gca().ticklabel_format(useOffset=False)
        if legends:  # FOR OLD MATPLOTLIB COMPATIBILITY
            plt.savefig(dest_file, bbox_extra_artists=(lgd,), bbox_inches="tight")
        else:
            plt.savefig(dest_file)
        plt.close()

    elif slice_dim == 2:

        if incl_uncert:
            x1, x2, nu = _get_2d_arrays(coords, settings, nu)
            _plot_2d_contours(x1, x2, nu, label=r'$\nu(x)$')
            _plot_2d_labels_settings(axis_labels)
            plt.tight_layout()
            plt.savefig(dest_file[:-4] + "_uncert.png")
            plt.close()

        x1, x2, mu = _get_2d_arrays(coords, settings, mu)
        _plot_2d_contours(x1, x2, mu, label=r'$\mu(x)$')

        lo = False
        lo |= _plot_2d_acqs(acqs, settings)
        lo |= _plot_2d_minima(minima, settings)
        lo |= _plot_2d_xnext(xnext, settings)
        lo |= _plot_2d_xhat(xhat, settings)
        _plot_2d_paths(paths, settings)

        _plot_2d_labels_settings(axis_labels)
        if legends and lo:
            lgd = plt.legend(
                bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
                loc=3,
                ncol=4,
                mode="expand",
                borderaxespad=0.0,
                prop={"size": 20},
            )
        if legends:  # FOR OLD MATPLOTLIB COMPATIBILITY
            plt.savefig(dest_file, bbox_extra_artists=(lgd,), bbox_inches="tight")
        else:
            plt.savefig(dest_file)
        plt.close()
    else:
        raise TypeError("ERROR: Model plot only applicable up to 2D (slices)")


def plot_acq_func(
    settings,
    dest_file,
    acqf_data,
    acqs=None,
    xhat=None,
    xnext=None,
    axis_labels=None,
    legends=True,
):
    """
    Plots a (max 2D) slice of the acquisition function. Can include also
    acquisitions, xhat and xnext if those are given.
    """
    if xnext is not None:
        xnext = shape_consistent_X(xnext, settings.dim)
    coords = acqf_data[:, : settings.dim]
    af = acqf_data[:, -1]
    slice_dim = (
        1 if settings["pp_model_slice"][0] == settings["pp_model_slice"][1] else 2
    )
    if axis_labels is None:
        axis_labels = [
            r"$x_%i$" % (settings["pp_model_slice"][0] + 1),
            r"$x_%i$" % (settings["pp_model_slice"][1] + 1),
        ]

    if slice_dim == 1:
        x1 = coords[:, settings["pp_model_slice"][0]]
        plt.plot(x1, af, "k", linewidth=5)

        if acqs is not None:
            x1 = acqs[:, settings["pp_model_slice"][0]]
            y = acqs[:, -1]
            plt.scatter(
                x1,
                y,
                s=200,
                linewidth=6,
                facecolors="none",
                edgecolors="brown",
                label="acqs",
            )

        if xhat is not None:
            plt.axvline(
                xhat[settings["pp_model_slice"][0]],
                color="red",
                linewidth=5,
                label=r"$x_{\mathrm{glmin}}$",
            )

        if xnext is not None:
            for xn in xnext:
                plt.axvline(
                    xn[settings["pp_model_slice"][0]],
                    color="green",
                    linewidth=5,
                    label=r"$x_{\mathrm{next}}$",
                    linestyle="dashed",
                )

        plt.xlim(
            min(coords[:, settings["pp_model_slice"][0]]),
            max(coords[:, settings["pp_model_slice"][0]]),
        )
        yd = max(af) - min(af)
        plt.ylim(min(af) - 0.1 * yd, max(af) + 0.1 * yd)
        plt.xlabel(axis_labels[0], size=24)
        plt.ylabel(r"$acqfn(x)$", size=24)
        if legends:
            lgd = plt.legend(
                bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
                loc=3,
                ncol=4,
                mode="expand",
                borderaxespad=0.0,
                prop={"size": 20},
            )
        plt.gcf().set_size_inches(10, 8)
        plt.gca().tick_params(labelsize=18)
        plt.gca().ticklabel_format(useOffset=False)
        if legends:  # FOR OLD MATPLOTLIB COMPATIBILITY
            plt.savefig(dest_file, bbox_extra_artists=(lgd,), bbox_inches="tight")
        else:
            plt.savefig(dest_file)
        plt.close()

    elif slice_dim == 2:
        x1, x2, af = _get_2d_arrays(coords, settings, af)
        _plot_2d_contours(x1, x2, af, label=r'$af(x)$')

        lo = False
        lo |= _plot_2d_acqs(acqs, settings)
        lo |= _plot_2d_xnext(xnext, settings)
        lo |= _plot_2d_xhat(xhat, settings)

        _plot_2d_labels_settings(axis_labels)
        if legends and lo:
            lgd = plt.legend(
                bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
                loc=3,
                ncol=4,
                mode="expand",
                borderaxespad=0.0,
                prop={"size": 20},
            )
        if legends and lo:  # FOR OLD MATPLOTLIB COMPATIBILITY
            plt.savefig(dest_file, bbox_extra_artists=(lgd,), bbox_inches="tight")
        else:
            plt.savefig(dest_file)
        plt.close()
    else:
        raise TypeError(
            "ERROR: Acquisition function plot only" + " applicable up to 2D (slices)"
        )


def plot_truef(settings, dest_file, truef_data, axis_labels=None):
    """
    Plots a (max 2D) slice of the true function.
    """
    coords = truef_data[:, : settings.dim]
    tf = truef_data[:, -1]
    slice_dim = (
        1 if settings["pp_model_slice"][0] == settings["pp_model_slice"][1] else 2
    )
    if axis_labels is None:
        axis_labels = [
            r"$x_%i$" % (settings["pp_model_slice"][0] + 1),
            r"$x_%i$" % (settings["pp_model_slice"][1] + 1),
        ]

    if slice_dim == 1:
        x1 = coords[:, settings["pp_model_slice"][0]]
        plt.plot(x1, tf, "k", linewidth=5)
        plt.xlim(min(x1), max(x1))
        plt.xlabel(axis_labels[0], size=24)
        plt.ylabel(r"$f(x)$", size=24)
        plt.title("True function", size=20)
        plt.gcf().set_size_inches(10, 8)
        plt.gca().tick_params(labelsize=18)
        plt.gca().ticklabel_format(useOffset=False)
        plt.tight_layout()
        plt.savefig(dest_file)
        plt.close()

    elif slice_dim == 2:
        x1, x2, tf = _get_2d_arrays(coords, settings, tf,
                                    npts=settings["pp_truef_npts"])
        _plot_2d_contours(x1, x2, tf, label=r'$f(x)$')
        _plot_2d_labels_settings(axis_labels)
        plt.title("True function", size=20)
        plt.tight_layout()
        plt.savefig(dest_file)
        plt.close()
    else:
        raise TypeError(
            "ERROR: True function plot only" + " applicable up to 2D (slices)"
        )


def _get_2d_arrays(coords, settings, data, *, npts=None):
    if npts is None:
        npts = settings["pp_model_slice"][2]
    # x1 and x2 here are like from meshgrid()
    x1 = coords[:, settings["pp_model_slice"][0]].reshape(npts, npts)
    x2 = coords[:, settings["pp_model_slice"][1]].reshape(npts, npts)
    reshaped_data = data.reshape(npts, npts)
    return x1, x2, reshaped_data


def _plot_2d_contours(x1, x2, data, *, label='data'):
    plt.contour(x1, x2, data, 25, colors='k')
    plt.pcolormesh(x1, x2, data, cmap='viridis', shading='gouraud',
                   rasterized=True)
    cbar = plt.colorbar()
    cbar.set_label(label=label, size=24)
    cbar.ax.tick_params(labelsize=18)
    plt.xlim(x1.min(), x1.max())
    plt.ylim(x2.min(), x2.max())


def _plot_2d_xhat(xhat, settings):
    if xhat is None:
        return False
    plt.plot(
        xhat[settings["pp_model_slice"][0]],
        xhat[settings["pp_model_slice"][1]],
        "r*",
        markersize=26,
        label=r"$x_{\mathrm{glmin}}$",
        zorder=10,
    )
    return True


def _plot_2d_acqs(acqs, settings):
    if acqs is None:
        return False
    x1 = acqs[:, settings["pp_model_slice"][0]]
    x2 = acqs[:, settings["pp_model_slice"][1]]
    sz = np.linspace(200, 500, len(x1))
    lw = np.linspace(3, 8, len(x1))
    plt.scatter(
        x1[0],
        x2[0],
        s=sz[int(len(x1) / 2.0)],
        linewidth=lw[int(len(x1) / 2.0)],
        facecolors="none",
        edgecolors="magenta",
        label="acqs",
        zorder=10,
    )
    plt.scatter(
        x1,
        x2,
        s=sz,
        linewidths=lw,
        facecolors="none",
        edgecolors="magenta",
        zorder=10,
    )
    return True


def _plot_2d_xnext(xnext, settings):
    if xnext is None:
        return False
    plt.plot(
        xnext[:, settings["pp_model_slice"][0]],
        xnext[:, settings["pp_model_slice"][1]],
        "b^",
        markersize=26,
        label=r"$x_{next}$",
        zorder=10,
    )
    return True


def _plot_2d_minima(minima, settings):
    if minima is None:
        return False
    x1 = minima[:, settings["pp_model_slice"][0]]
    x2 = minima[:, settings["pp_model_slice"][1]]
    plt.scatter(
        x1,
        x2,
        s=350,
        linewidth=6,
        facecolors="none",
        edgecolors="navajowhite",
        label="minima",
        zorder=10,
    )
    return True


def _plot_2d_paths(paths, settings):
    if paths is None:
        return
    threshold = min(settings["bounds"][:, 1] - settings["bounds"][:, 0]) / 2
    for path in paths:
        start = 0
        stop = 1
        while True:
            if (
                stop == path.crds.shape[0] - 1
                or np.linalg.norm(path.crds[stop, :] - path.crds[stop - 1, :])
                > threshold
            ):
                plt.plot(
                    path.crds[range(start, stop), 0],
                    path.crds[range(start, stop), 1],
                    linewidth=3.0,
                    color="red",
                )
                if stop == path.crds.shape[0] - 1:
                    break
                else:
                    start = stop
            stop += 1


def _plot_2d_labels_settings(axis_labels):
    # XXX one could use mpl.rcParams or styles to set these
    plt.xlabel(axis_labels[0], size=24)
    plt.ylabel(axis_labels[1], size=24)
    plt.gcf().set_size_inches(10, 8)
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)
